// The CERN-OHL-P v2 License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// This source describes Open Hardware and is licensed under the CERN-OHL-P v2.
// You may redistribute and modify this documentation and make products using it under the terms of
// the CERN-OHL-P v2 (https:/cern.ch/cern-ohl). This documentation is distributed WITHOUT ANY
// EXPRESS OR IMPLIED WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A
// PARTICULAR PURPOSE. Please see the CERN-OHL-P v2 for applicable conditions.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
module apb3_slave # (
		parameter ADDR_WIDTH = 16,
		parameter DATA_WIDTH = 32,
		parameter LVDS_TEST_NUMS = 5,
		parameter SLED_ADDR_WIDTH = 2,								// Max 7bit	128LEDs
		parameter SLED_DATA_WIDTH = 24								// Fixed 24bit
	) (
		input CLOCK,
		input RESET_N,

		// INTERRUPT
		output INTERRUPT,

		// APB3
		input [ADDR_WIDTH-1:0] PADDR,
		input PSEL,
		input PENABLE,
		output PREADY,
		input PWRITE,
		input [DATA_WIDTH-1:0] PWDATA,
		output [DATA_WIDTH-1:0] PRDATA,
		output PSLVERROR,

		// System information
		input [DATA_WIDTH-1:0] RTL_VERSION_IN,
		input [DATA_WIDTH-1:0] RTL_DATE_IN,

		// Power monitor and control
		input PWR_SW_N_IN,
		input VIN_MON_N_IN,
		input VBUS_MON_N_IN,
		output PWR_CNT_OUT,
		output USB_POW_DIS_OUT,
		output ISET_1R0A_N_OUT,
		output ISET_1R5A_N_OUT,

		// USB TYPE-C
		input TYPEC_SWMONI_IN,
		input [1:0] TYPEC_TCC_IN,

		// USB VCP CP2102N
		output USB_VCP_RTS_OUT,
		input USB_VCP_CTS_IN,
		input [1:0] USB_CHR_IN,
		input USB_CHREN_IN,
		input USB_SUSPEND_IN,

		// SD Card
		input SD_DET_N_IN,

		// ADC, DAC
		output AREF_EXT_SEL_N_OUT,

		// LED
		output PWR_LED_R_N_OUT,
		output PWR_LED_G_N_OUT,
		output PWR_LED_B_N_OUT,
		output [SLED_ADDR_WIDTH-1:0] SER_LED_ADDR_OUT,
		output [SLED_DATA_WIDTH-1:0] SER_LED_DATA_OUT,
		output SER_LED_WRITE_OUT,
		output SER_LED_SET_OUT,
		input SER_LED_BUSY_IN,

		// LVDS
		output LVDS_TXPLL_RST_N_OUT,
		input LVDS_TXPLL_LOCKED_IN,
		output LVDS_RXPLL_RST_N_OUT,
		input LVDS_RXPLL_LOCKED_IN,
		output [LVDS_TEST_NUMS-1:0] LVDS_TEST_RUN_OUT,
		input [LVDS_TEST_NUMS-1:0] LVDS_TEST_RX_IN,
		input [LVDS_TEST_NUMS-1:0] LVDS_TEST_ERR_IN
	);

	localparam STATE_WIDTH = 2;
	localparam [STATE_WIDTH-1:0] STATE_IDLE = 0,
								 STATE_SETUP = 1,
								 STATE_ACCESS = 2;

	localparam [ADDR_WIDTH-1:0] VERSION_REG_ADDR = 'h0,
								DATE_REG_ADDR = 'h4,
								CONTROL_REG_ADDR = 'h10,
								STATUS_REG_ADDR = CONTROL_REG_ADDR,
								STATUS_CHG_REG_ADDR = 'h14,
								STATUS_CLR_CHG_REG_ADDR = STATUS_CHG_REG_ADDR,
								CONTROL_SET_REG_ADDR = 'h18,
								CONTROL_CLR_REG_ADDR = 'h1C,
								INT_UPEG_REG_ADDR = 'h20,
								INT_DNEG_REG_ADDR = 'h24,
								INT_REG_ADDR = 'h28,
								INT_CLR_REG_ADDR = INT_REG_ADDR,
								SLED_REG_ADDR = 'h30;

	reg [STATE_WIDTH-1:0] bus_state_reg;
	reg [DATA_WIDTH-1:0] control_reg;
	reg [DATA_WIDTH-1:0] status_reg;
	reg [SLED_ADDR_WIDTH-1:0] sled_addr_reg;
	reg [SLED_DATA_WIDTH-1:0] sled_data_reg;
	reg sled_write_reg;
	reg sled_set_reg;
	reg [DATA_WIDTH-1:0] int_upeg_reg;
	reg [DATA_WIDTH-1:0] int_dneg_reg;
	reg [DATA_WIDTH-1:0] chg_reg;
	reg [DATA_WIDTH-1:0] int_reg;
	reg [DATA_WIDTH-1:0] read_data_reg;
	reg slave_ready_reg;

	wire [DATA_WIDTH-1:0] status;
	wire act_write;
	wire act_read;
	wire pwr_sw_n;
	wire vin_mon_n;
	wire vbus_mon_n;
	wire typec_swmoni;
	wire [1:0] typec_tcc;
	wire usb_vcp_cts;
	wire [1:0] usb_chr;
	wire usb_chren;
	wire usb_suspend;
	wire sd_det_n;
	wire lvds_txpll_locked;
	wire lvds_rxpll_locked;
	wire [LVDS_TEST_NUMS-1:0] lvds_test_rx;
	wire [LVDS_TEST_NUMS-1:0] lvds_test_err;
	wire lvds_test_rx_status;
	wire lvds_test_err_status;

	assign INTERRUPT = |int_reg;

	assign PWR_CNT_OUT = control_reg[3];
	assign USB_POW_DIS_OUT = control_reg[4];
	assign ISET_1R0A_N_OUT = control_reg[5];
	assign ISET_1R5A_N_OUT = control_reg[6];
	assign USB_VCP_RTS_OUT = control_reg[10];
	assign AREF_EXT_SEL_N_OUT = control_reg[17];
	assign PWR_LED_R_N_OUT = control_reg[18];
	assign PWR_LED_G_N_OUT = control_reg[19];
	assign PWR_LED_B_N_OUT = control_reg[20];
	assign LVDS_TXPLL_RST_N_OUT = control_reg[21];
	assign LVDS_RXPLL_RST_N_OUT = control_reg[23];
	assign LVDS_TEST_RUN_OUT = control_reg[29:25];

	assign lvds_test_rx_status = (|control_reg[29:25]) & (&(lvds_test_rx | ~control_reg[29:25]));
	assign lvds_test_err_status = |lvds_test_err;

	s_ff # (
		.CLOCK_POL(1),
		.RESET_POL(0),
		.ARESET(1),
		.DATA_WIDTH(12),
		.SHIFT_NUMS(2)
	) s_ff_u1 (
		.CLOCK(CLOCK), 
		.RESET(RESET_N), 
		.IN_DATA({SD_DET_N_IN, USB_SUSPEND_IN, USB_CHREN_IN, USB_CHR_IN, USB_VCP_CTS_IN, TYPEC_TCC_IN, TYPEC_SWMONI_IN, VBUS_MON_N_IN, VIN_MON_N_IN, PWR_SW_N_IN}),
		.OUT_DATA({sd_det_n, usb_suspend, usb_chren, usb_chr, usb_vcp_cts, typec_tcc, typec_swmoni, vbus_mon_n, vin_mon_n, pwr_sw_n})
	);

	s_ff # (
		.CLOCK_POL(1),
		.RESET_POL(0),
		.ARESET(1),
		.DATA_WIDTH(12),
		.SHIFT_NUMS(2)
	) s_ff_u2 (
		.CLOCK(CLOCK), 
		.RESET(RESET_N), 
		.IN_DATA({LVDS_TEST_ERR_IN, LVDS_TEST_RX_IN, LVDS_RXPLL_LOCKED_IN, LVDS_TXPLL_LOCKED_IN}),
		.OUT_DATA({lvds_test_err, lvds_test_rx, lvds_rxpll_locked, lvds_txpll_locked})
	);

	assign status = {
		// LVDS
		lvds_test_err_status,		// [31]
		lvds_test_rx_status,		// [30]
		control_reg[29:25],			// LVDS_TEST_OUT
		lvds_rxpll_locked,			// [24]
		control_reg[23],			// LVDS_RXPLL_RST_N_OUT
		lvds_txpll_locked,			// [22]
		control_reg[21],			// LVDS_TXPLL_RST_N_OUT
		// Power LED
		control_reg[20:18],			// {PWR_LED_B_N_OUT, PWR_LED_G_N_OUT, PWR_LED_R_N_OUT}
		// ADC, DAC
		control_reg[17],			// AREF_EXT_SEL_N_OUT
		// SD Card
		sd_det_n,					// [16]
		// USB VCP CP2102N
		usb_suspend,				// [15]
		usb_chren,					// [14]
		usb_chr,					// [13:12]
		usb_vcp_cts,				// [11]
		control_reg[10],			// USB_VCP_RTS_OUT
		// USB TYPE-C
		typec_tcc,					// [9:8]
		typec_swmoni,				// [7]
		// Power monitor
		control_reg[6:3],			// {ISET_1R5A_N_OUT, ISET_1R0A_N_OUT, USB_POW_DIS, PWR_CNT}
		vbus_mon_n,					// [2]
		vin_mon_n,					// [1]
		pwr_sw_n					// [0]
	};

	assign SER_LED_ADDR_OUT = sled_addr_reg[SLED_ADDR_WIDTH-1:0];
	assign SER_LED_DATA_OUT = sled_data_reg[SLED_DATA_WIDTH-1:0];
	assign SER_LED_WRITE_OUT = sled_write_reg;
	assign SER_LED_SET_OUT = sled_set_reg;

	assign act_write = (PWRITE) && (bus_state_reg == STATE_ACCESS);
	assign act_read = (!PWRITE) && (bus_state_reg == STATE_ACCESS);
	assign PSLVERROR = 1'b0; 
	assign PRDATA = read_data_reg;
	assign PREADY = slave_ready_reg && (bus_state_reg !== STATE_IDLE);

	always @(posedge CLOCK or negedge RESET_N)
	begin
		if (!RESET_N) 
			bus_state_reg <= STATE_IDLE; 
		else begin
			case (bus_state_reg)
				STATE_IDLE : bus_state_reg <= (PSEL && !PENABLE)? STATE_SETUP : STATE_IDLE;
				STATE_SETUP : bus_state_reg <= (PSEL && PENABLE)? STATE_ACCESS : STATE_IDLE;
				STATE_ACCESS : bus_state_reg <= (PREADY)? STATE_IDLE : STATE_ACCESS;
				default : bus_state_reg <= STATE_IDLE;
			endcase
		end
	end

	always@ (posedge CLOCK or negedge RESET_N)
	begin
		if (!RESET_N) 
			slave_ready_reg = 1'b0;
		else
			slave_ready_reg <= act_write || act_read;
	end

	always@ (posedge CLOCK or negedge RESET_N)
	begin
		if (!RESET_N) begin
			control_reg[2:0] <= 3'b000;
			control_reg[3] <= 1'b1;			// PWR_CNT_OUT -> PWR_CNT : ON
			control_reg[4] <= 1'b0;			// USB_POW_DIS_OUT -> USB_POW_DIS : OFF
			control_reg[5] <= 1'b1;			// ISET_1R0A_N_OUT -> ISET 1R0A : OFF
			control_reg[6] <= 1'b1;			// ISET_1R5A_N_OUT -> ISET 1R5A : OFF
			control_reg[9:7] <= 3'b000;
			control_reg[10] <= 1'b1;		// USB_VCP_RTS_OUT -> RTS : OFF 
			control_reg[16:11] <= 6'b000000;
			control_reg[17] <= 1'b1;		// AREF_EXT_SEL_N_OUT -> AREF = INT
			control_reg[18] <= 1'b0;		// PWR_LED_R_N_OUT -> LED : ON
			control_reg[19] <= 1'b1;		// PWR_LED_G_N_OUT -> LED : OFF
			control_reg[20] <= 1'b1;		// PWR_LED_B_N_OUT -> LED : OFF
			control_reg[21] <= 1'b0;		// LVDS_TXPLL_RST_N_OUT -> TXPLL : RESET
			control_reg[22] <= 1'b0;
			control_reg[23] <= 1'b0;		// LVDS_RXPLL_RST_N_OUT -> RXPLL : RESET
			control_reg[24] <= 1'b0;
			control_reg[29:25] <= 5'b00000;	// LVDS_TEST_RUN_OUT -> LVDS TEST : OFF
			control_reg[31:30] <= 2'b00;
		end
		else begin
			if (act_write) begin
				if (PADDR == CONTROL_REG_ADDR)
					control_reg <= PWDATA;
				else if (PADDR == CONTROL_SET_REG_ADDR)
					control_reg <= control_reg | PWDATA;
				else if (PADDR == CONTROL_CLR_REG_ADDR)
					control_reg <= control_reg & ~PWDATA;
			end
		end
	end

	always@ (posedge CLOCK or negedge RESET_N)
	begin
		if (!RESET_N)
			int_upeg_reg <= {DATA_WIDTH{1'b0}};
		else begin
			if ((act_write) & (PADDR == INT_UPEG_REG_ADDR))
				int_upeg_reg <= PWDATA;
		end
	end

	always@ (posedge CLOCK or negedge RESET_N)
	begin
		if (!RESET_N)
			int_dneg_reg <= {DATA_WIDTH{1'b0}};
		else begin
			if ((act_write) & (PADDR == INT_DNEG_REG_ADDR))
				int_dneg_reg <= PWDATA;
		end
	end

	always@ (posedge CLOCK or negedge RESET_N)
	begin
		if (!RESET_N)
			status_reg <= {DATA_WIDTH{1'b0}};
		else begin
			status_reg <= status;
		end
	end

	always@ (posedge CLOCK or negedge RESET_N)
	begin
		if (!RESET_N)
			chg_reg <= {DATA_WIDTH{1'b1}};
		else begin
			for (integer i = 0; i < DATA_WIDTH; i = i + 1) begin
				if (~chg_reg[i]) begin
					if (status_reg[i] ^ status[i]) 
						chg_reg[i] <= 1'b1;
				end
				else if (act_write) begin
					if (PADDR == STATUS_CLR_CHG_REG_ADDR) begin
						if (PWDATA[i])
							chg_reg[i] <= 1'b0;
					end
				end
			end
		end
	end

	always@ (posedge CLOCK or negedge RESET_N)
	begin
		if (!RESET_N)
			int_reg <= {DATA_WIDTH{1'b0}};
		else begin
			for (integer i = 0; i < DATA_WIDTH; i = i + 1) begin
				if (~int_reg[i]) begin
					if (int_upeg_reg[i] && ~status_reg[i] && status[i]) 
						int_reg[i] <= 1'b1;
					else if (int_dneg_reg[i] && status_reg[i] && ~status[i]) 
						int_reg[i] <= 1'b1;
				end
				else if (act_write) begin
					if (PADDR == INT_CLR_REG_ADDR) begin
						if (PWDATA[i])
							int_reg[i] <= 1'b0;
					end
				end
			end
		end
	end

	always@ (posedge CLOCK or negedge RESET_N)
	begin
		if (!RESET_N)
			sled_addr_reg <= {SLED_ADDR_WIDTH{1'b0}};
		else begin
			if ((act_write) & (PADDR == SLED_REG_ADDR))
				sled_addr_reg <= PWDATA[SLED_ADDR_WIDTH+SLED_DATA_WIDTH-1:SLED_DATA_WIDTH];
		end
	end

	always@ (posedge CLOCK or negedge RESET_N)
	begin
		if (!RESET_N)
			sled_data_reg <= {SLED_DATA_WIDTH{1'b0}};
		else begin
			if ((act_write) & (PADDR == SLED_REG_ADDR))
				sled_data_reg <= PWDATA[SLED_DATA_WIDTH-1:0];
		end
	end

	always@ (posedge CLOCK or negedge RESET_N)
	begin
		if (!RESET_N)
			sled_write_reg <= 1'b0;
		else begin
			if ((act_write) & (PADDR == SLED_REG_ADDR))
				sled_write_reg <= 1'b1;
			else
				sled_write_reg <= 1'b0;
		end
	end

	always@ (posedge CLOCK or negedge RESET_N)
	begin
		if (!RESET_N)
			sled_set_reg <= 1'b0;
		else begin
			if ((act_write) & (PADDR == SLED_REG_ADDR))
				sled_set_reg <= PWDATA[DATA_WIDTH-1];
			else
				sled_set_reg <= 1'b0;
		end
	end

	always@ (posedge CLOCK or negedge RESET_N)
	begin
		if(!RESET_N)
			read_data_reg <= {DATA_WIDTH{1'b0}};
		else begin
			if(act_read) begin
				case (PADDR) 
					VERSION_REG_ADDR : read_data_reg <= RTL_VERSION_IN;
					DATE_REG_ADDR : read_data_reg <= RTL_DATE_IN;
					STATUS_REG_ADDR : read_data_reg <= status;
					STATUS_CHG_REG_ADDR : read_data_reg <= chg_reg;
					INT_UPEG_REG_ADDR : read_data_reg <= int_upeg_reg;
					INT_DNEG_REG_ADDR : read_data_reg <= int_dneg_reg;
					INT_REG_ADDR : read_data_reg <= int_reg;
					SLED_REG_ADDR : read_data_reg <= {SER_LED_BUSY_IN, {(DATA_WIDTH-SLED_ADDR_WIDTH-SLED_DATA_WIDTH-1){1'b0}}, sled_addr_reg, sled_data_reg};
					default : read_data_reg <= {DATA_WIDTH{1'b0}};
				endcase
			end
		end
	end

endmodule
