// The CERN-OHL-P v2 License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// This source describes Open Hardware and is licensed under the CERN-OHL-P v2.
// You may redistribute and modify this documentation and make products using it under the terms of
// the CERN-OHL-P v2 (https:/cern.ch/cern-ohl). This documentation is distributed WITHOUT ANY
// EXPRESS OR IMPLIED WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A
// PARTICULAR PURPOSE. Please see the CERN-OHL-P v2 for applicable conditions.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
`timescale 1ps / 1ps

module sim;

	// Parameters
	parameter CLK_CYCLE = 20000;							// 50MHz
	parameter LVDS_SCLK_CYCLE = 16000;						// 62.5MHz
	parameter ADDR_WIDTH = 16;
	parameter DATA_WIDTH = 32;
	parameter LVDS_TEST_NUMS = 5;
	parameter SLED_ADDR_WIDTH = 2;
	parameter SLED_DATA_WIDTH = 24;							// Fixed 24bit
	parameter [31:0] RTL_VERSION = "010A";
	parameter [31:0] RTL_DATE = {16'd2023, 8'd3, 8'd31};
	parameter TIMEOUT = 10;

	reg CLOCK = 1'b0;
	reg RESET_N = 1'b0;

	wire INTERRUPT;

	// APB3
	reg [ADDR_WIDTH-1:0] PADDR = {ADDR_WIDTH{1'b0}};
	reg PSEL = 1'b0;
	reg PENABLE = 1'b0;
	wire PREADY;
	reg PWRITE = 1'b0;
	reg [DATA_WIDTH-1:0] PWDATA = {DATA_WIDTH{1'b0}};
	wire [DATA_WIDTH-1:0] PRDATA;
	wire PSLVERROR;

	// System information
	reg [DATA_WIDTH-1:0] RTL_VERSION_IN = RTL_VERSION;
	reg [DATA_WIDTH-1:0] RTL_DATE_IN = RTL_DATE;

	// Power monitor and control
	reg PWR_SW_N_IN = 1'b1;
	reg VIN_MON_N_IN = 1'b1;
	reg VBUS_MON_N_IN = 1'b1;
	wire PWR_CNT_OUT;
	wire USB_POW_DIS_OUT;
	wire ISET_1R0A_N_OUT;
	wire ISET_1R5A_N_OUT;

	// USB TYPE-C
	reg TYPEC_SWMONI_IN = 1'b0;
	reg [1:0] TYPEC_TCC_IN = 2'b00;

	// USB VCP CP2102N
	wire USB_VCP_RTS_OUT;
	reg USB_VCP_CTS_IN = 1'b1;
	reg [1:0] USB_CHR_IN = 2'b00;
	reg USB_CHREN_IN = 1'b0;
	reg USB_SUSPEND_IN = 1'b0;

	// SD Card
	reg SD_DET_N_IN = 1'b1;

	// ADC, DAC
	wire AREF_EXT_SEL_N_OUT;

	// LED
	wire PWR_LED_R_N_OUT;
	wire PWR_LED_G_N_OUT;
	wire PWR_LED_B_N_OUT;
	wire [SLED_ADDR_WIDTH-1:0] SER_LED_ADDR_OUT;
	wire [SLED_DATA_WIDTH-1:0] SER_LED_DATA_OUT;
	wire SER_LED_WRITE_OUT;
	wire SER_LED_SET_OUT;
	reg SER_LED_BUSY_IN = 1'b0;

	// LVDS
	wire LVDS_TXPLL_RST_N_OUT;
	reg LVDS_TXPLL_LOCKED_IN = 1'b0;
	wire LVDS_RXPLL_RST_N_OUT;
	reg LVDS_RXPLL_LOCKED_IN = 1'b0;
	wire [LVDS_TEST_NUMS-1:0] LVDS_TEST_RUN_OUT;
	reg [LVDS_TEST_NUMS-1:0] LVDS_TEST_RX_IN = {LVDS_TEST_NUMS{1'b0}};
	reg [LVDS_TEST_NUMS-1:0] LVDS_TEST_ERR_IN = {LVDS_TEST_NUMS{1'b0}};

	// Test wire
	reg [DATA_WIDTH-1:0] read_ver = {DATA_WIDTH{1'b0}};
	reg [DATA_WIDTH-1:0] read_date = {DATA_WIDTH{1'b0}};
	reg [DATA_WIDTH-1:0] control_data = {DATA_WIDTH{1'b0}};
	reg [DATA_WIDTH-1:0] status_chg = {DATA_WIDTH{1'b0}};
	reg [DATA_WIDTH-1:0] int_data = {DATA_WIDTH{1'b0}};
	reg [DATA_WIDTH-1:0] sled_data = {DATA_WIDTH{1'b0}};
	reg timeout_error = 1'b0;

	integer i = 0;

	apb3_slave dut1 (
		.*
	);

	initial begin
		$dumpfile("outflow/apb3_slave.vcd");
		$dumpvars(0, sim);
	end

	always #(CLK_CYCLE / 2) CLOCK = ~CLOCK;

	task task_delay (
			input integer value
		);
		begin
			repeat (value)
				@(posedge CLOCK);
			#1;
		end
	endtask

	task apb3_write (
			input [ADDR_WIDTH-1:0] addr,
			input [DATA_WIDTH-1:0] data
		);
		begin
			@(posedge CLOCK)
			#1	PADDR <= addr;
				PWDATA <= data;
				PWRITE <= 1'b1;
				PSEL <= 1'b1;
			@(posedge CLOCK)
			#1	PENABLE <= 1'b1;
			@(posedge CLOCK)
			for (integer t = 0; (t < TIMEOUT) && (!PREADY); t = t + 1) begin
				@(posedge CLOCK);
			end
			if (!PREADY)
				timeout_error <= 1'b1;
			#1	PSEL <= 1'b0;
				PENABLE <= 1'b0;
		end
	endtask

	task apb3_read (
			input [ADDR_WIDTH-1:0] addr,
			output [DATA_WIDTH-1:0] data
		);
		begin
			@(posedge CLOCK)
			#1	PADDR <= addr;
				PWRITE <= 1'b0;
				PSEL <= 1'b1;
			@(posedge CLOCK)
			#1	PENABLE <= 1'b1;
			@(posedge CLOCK)
			for (integer t = 0; (t < TIMEOUT) && (!PREADY); t = t + 1) begin
				@(posedge CLOCK);
			end
			if (!PREADY)
				timeout_error <= 1'b1;
			data <= PRDATA;
			#1	PSEL <= 1'b0;
				PENABLE <= 1'b0;
		end
	endtask

	initial begin
		#(CLK_CYCLE * 5.25);
		RESET_N = 1'b1;

		task_delay(100);

		apb3_read(dut1.VERSION_REG_ADDR, read_ver);			task_delay(100);
		apb3_read(dut1.DATE_REG_ADDR, read_date);			task_delay(100);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(100);
		apb3_read(dut1.SLED_REG_ADDR, sled_data);			task_delay(100);

		apb3_write(dut1.CONTROL_REG_ADDR, control_data);	task_delay(100);

		for (integer i = 0; i < DATA_WIDTH; i = i + 1) begin
			apb3_write(dut1.CONTROL_REG_ADDR, control_data ^ ({{DATA_WIDTH-1{1'b0}},1'b1} << i));
			task_delay(10);
			apb3_read(dut1.STATUS_REG_ADDR, control_data);
			task_delay(10);
		end

		task_delay(100);

		for (integer i = 0; i < DATA_WIDTH; i = i + 1) begin
			apb3_write(dut1.SLED_REG_ADDR, {{DATA_WIDTH-1{1'b0}},1'b1} << i);
			task_delay(10);
			apb3_read(dut1.SLED_REG_ADDR, sled_data);
			task_delay(10);
		end

		SER_LED_BUSY_IN = 1'b1;
		apb3_read(dut1.SLED_REG_ADDR, sled_data);

		// Power monitor and control
		PWR_SW_N_IN = 1'b0;									task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		VIN_MON_N_IN = 1'b0;								task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		VBUS_MON_N_IN = 1'b0;								task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);

		// USB TYPE-C
		TYPEC_SWMONI_IN = 1'b1;								task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		TYPEC_TCC_IN[0] = 1'b1;								task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		TYPEC_TCC_IN[1] = 1'b1;								task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);

		// USB VCP CP2102N
		USB_VCP_CTS_IN = 1'b0;								task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		USB_CHR_IN[0] = 1'b1;								task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		USB_CHR_IN[1] = 1'b1;								task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		USB_CHREN_IN = 1'b1;								task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		USB_SUSPEND_IN = 1'b1;								task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);

		// SD Card
		apb3_write(dut1.INT_DNEG_REG_ADDR, 32'h00010000);	task_delay(100);

		SD_DET_N_IN = 1'b0;									task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);

		apb3_read(dut1.INT_REG_ADDR, int_data);				task_delay(10);
		apb3_write(dut1.INT_CLR_REG_ADDR, 32'h00010000);	task_delay(100);
		apb3_read(dut1.INT_REG_ADDR, int_data);				task_delay(10);

		apb3_write(dut1.INT_DNEG_REG_ADDR, 32'h00000000);	task_delay(100);
		apb3_write(dut1.INT_UPEG_REG_ADDR, 32'h00010000);	task_delay(100);

		SD_DET_N_IN = 1'b1;									task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);

		apb3_read(dut1.INT_REG_ADDR, int_data);				task_delay(10);
		apb3_write(dut1.INT_CLR_REG_ADDR, 32'hFFFEFFFF);	task_delay(100);
		apb3_read(dut1.INT_REG_ADDR, int_data);				task_delay(10);

		apb3_write(dut1.INT_CLR_REG_ADDR, 32'h00010000);	task_delay(100);
		apb3_read(dut1.INT_REG_ADDR, int_data);				task_delay(10);

		apb3_write(dut1.INT_UPEG_REG_ADDR, 32'h00000000);	task_delay(100);

		SD_DET_N_IN = 1'b0;									task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);

		// LVDS
		LVDS_TXPLL_LOCKED_IN = 1'b1;						task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		LVDS_RXPLL_LOCKED_IN = 1'b1;						task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		LVDS_TEST_RX_IN[0] = 1'b1;							task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		LVDS_TEST_RX_IN[1] = 1'b1;							task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		LVDS_TEST_RX_IN[2] = 1'b1;							task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		LVDS_TEST_RX_IN[3] = 1'b1;							task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		LVDS_TEST_RX_IN[4] = 1'b1;							task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);
		LVDS_TEST_RX_IN[0] = 1'b0;							task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);

		control_data[25] = 1'b0;	// LVDS_TEST_OUT[0]
		apb3_write(dut1.CONTROL_REG_ADDR, control_data);	task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);

		LVDS_TEST_ERR_IN[0] = 1'b1;							task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);

		LVDS_TEST_ERR_IN[0] = 1'b0;							task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);

		LVDS_TEST_ERR_IN[4] = 1'b1;							task_delay(10);
		apb3_read(dut1.STATUS_REG_ADDR, control_data);		task_delay(10);

		// Bit Control and change detect

		apb3_read(dut1.STATUS_CHG_REG_ADDR, status_chg);
		task_delay(10);

		apb3_write(dut1.STATUS_CLR_CHG_REG_ADDR, 32'hFFFFFFFF);
		task_delay(10);

		apb3_read(dut1.STATUS_CHG_REG_ADDR, status_chg);
		task_delay(10);

		apb3_write(dut1.CONTROL_SET_REG_ADDR, 32'h00000008);
		task_delay(10);

		apb3_read(dut1.STATUS_CHG_REG_ADDR, status_chg);
		task_delay(10);

		apb3_write(dut1.STATUS_CLR_CHG_REG_ADDR, 32'h00000008);
		task_delay(10);

		apb3_read(dut1.STATUS_CHG_REG_ADDR, status_chg);
		task_delay(10);

		apb3_write(dut1.CONTROL_CLR_REG_ADDR, 32'h00000008);
		task_delay(10);

		apb3_read(dut1.STATUS_CHG_REG_ADDR, status_chg);
		task_delay(10);

		apb3_write(dut1.STATUS_CLR_CHG_REG_ADDR, 32'h00000008);
		task_delay(10);

		apb3_read(dut1.STATUS_CHG_REG_ADDR, status_chg);
		task_delay(10);

		task_delay(100);
		$finish;
	end

endmodule
