// The CERN-OHL-P v2 License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// This source describes Open Hardware and is licensed under the CERN-OHL-P v2.
// You may redistribute and modify this documentation and make products using it under the terms of
// the CERN-OHL-P v2 (https:/cern.ch/cern-ohl). This documentation is distributed WITHOUT ANY
// EXPRESS OR IMPLIED WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A
// PARTICULAR PURPOSE. Please see the CERN-OHL-P v2 for applicable conditions.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
module counter # (
		parameter CLOCK_POL = 1,
		parameter RESET_POL = 0,
		parameter ARESET = 1,
		parameter CT_WIDTH = 4,
		parameter CONTINUE = 0,
		parameter CARRY_START = 0
	) (
		input CLOCK,
		input RESET,
		input [CT_WIDTH-1:0] IN_COUNT,
		input IN_COUNT_CLR,
		input IN_COUNT_RUN,
		input IN_COUNT_STB,
		input [CT_WIDTH-1:0] IN_COMP_LOW,
		input [CT_WIDTH-1:0] IN_COMP_HIGH,
		output [CT_WIDTH-1:0] OUT_COUNT,
		output OUT_CARRY,
		output OUT_COMP,
		output OUT_COUNT_STB
	);

	reg [CT_WIDTH-1:0] ct_reg;
	reg ct_stb_reg;
	reg carry_reg;

	wire clk, rst, a_rst, s_rst;
	wire [CT_WIDTH-1:0] next_ct;
	wire [CT_WIDTH-1:0] clear_ct;

	assign clk = (CLOCK_POL == 0)? ~CLOCK : CLOCK;
	assign rst = (RESET_POL == 0)? ~RESET : RESET;
	assign a_rst = (ARESET != 0)? rst : 1'b0;
	assign s_rst = (ARESET == 0)? rst : 1'b0;

	assign clear_ct = ((CARRY_START) && (CONTINUE))? IN_COUNT - {{CT_WIDTH-1{1'b0}}, 1'b1} : {CT_WIDTH{1'b0}};
	assign next_ct = ct_reg + {{CT_WIDTH-1{1'b0}}, 1'b1};

	assign OUT_COUNT = ct_reg;
	assign OUT_COUNT_STB = ct_stb_reg;
	assign OUT_CARRY = (carry_reg);
	assign OUT_COMP = (ct_reg <= IN_COMP_HIGH) && (ct_reg >= IN_COMP_LOW);

	always @(posedge clk or posedge a_rst) begin
		if (a_rst)
			ct_reg <= clear_ct;
		else if (s_rst)
			ct_reg <= clear_ct;
		else begin
			if (IN_COUNT_CLR)
				ct_reg <= clear_ct;
			else if ((IN_COUNT_RUN) && (IN_COUNT_STB)) begin
				if (CONTINUE) begin
					if (next_ct < IN_COUNT)
						ct_reg <= next_ct;
					else
						ct_reg <= {CT_WIDTH{1'b0}};
				end
				else begin
					if (ct_reg < IN_COUNT)
						ct_reg <= next_ct;
				end
			end
		end
	end

	always @(posedge clk or posedge a_rst) begin
		if (a_rst)
			ct_stb_reg <= 1'b0;
		else if (s_rst)
			ct_stb_reg <= 1'b0;
		else begin
			if ((!IN_COUNT_CLR) && (IN_COUNT_RUN) && (IN_COUNT_STB) && (IN_COUNT != {CT_WIDTH{1'b0}})) begin
				if (CONTINUE)
					ct_stb_reg <= 1'b1;
				else
					ct_stb_reg <= (ct_reg < IN_COUNT);
			end
			else
				ct_stb_reg <= 1'b0;
		end
	end

	always @(posedge clk or posedge a_rst) begin
		if (a_rst)
			carry_reg <= 1'b0;
		else if (s_rst)
			carry_reg <= 1'b0;
		else begin
			if ((IN_COUNT_CLR) || (!IN_COUNT_RUN))
				carry_reg <= 1'b0;
			else begin
				if ((IN_COUNT_STB) && (IN_COUNT != {CT_WIDTH{1'b0}})) begin
					if (CONTINUE)
						carry_reg <= (next_ct >= IN_COUNT);
					else
						carry_reg <= (next_ct >= IN_COUNT) && (ct_reg < IN_COUNT);
				end
				else
					carry_reg <= 1'b0;
			end
		end
	end

endmodule
