// The MIT License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software
// and associated documentation files (the “Software”), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge, publish, distribute,
// sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
// NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
#include <stdbool.h>

#include <bsp.h>

#include "interrupt.h"
#include "test.h"
#include "power.h"
#include "lib/apb3.h"

#define SLED_TEST_COUNT_MAX		0xFF
#define SLED_TEST_STEP_MAX		3
#define SLED_CYCLE_US			(1000*1000)

bool apb3_dispVerDateTest(test_state_t test_state)
{
	if (test_state == test_state_start) {
		apb3_dispVerDate();
	}

	return false;
}

bool apb3_sledTest(test_state_t test_state)
{
	#define UPCT		count
	#define DNCT		SLED_TEST_COUNT_MAX - count


	static int count = 0;
	static int step = 0;

	if (test_state == test_state_end) {
		SLED_BLK(1,0);
		SLED_BLK(2,1);
	    return false;
	}

	if (test_state == test_state_start) {
		count = 0;
		step = 0;
	}

	switch (step) {
	case 0:
		set_led((sled_t){.ad=1, .r=UPCT, .g=0x00, .b=0x00, .set=0});
		set_led((sled_t){.ad=2, .r=0x00, .g=UPCT, .b=0x00, .set=1});
		break;
	case 1:
		set_led((sled_t){.ad=1, .r=DNCT, .g=UPCT, .b=0x00, .set=0});
		set_led((sled_t){.ad=2, .r=0x00, .g=DNCT, .b=UPCT, .set=1});
		break;
	case 2:
		set_led((sled_t){.ad=1, .r=0x00, .g=DNCT, .b=UPCT, .set=0});
		set_led((sled_t){.ad=2, .r=UPCT, .g=0x00, .b=DNCT, .set=1});
		break;
	case 3:
		set_led((sled_t){.ad=1, .r=UPCT, .g=0x00, .b=DNCT, .set=0});
		set_led((sled_t){.ad=2, .r=DNCT, .g=UPCT, .b=0x00, .set=1});
		break;
	default:
		break;
	}

	if (count < SLED_TEST_COUNT_MAX) {
		count++;
	} else {
		count = 0;
		if (step < SLED_TEST_STEP_MAX) {
			step++;
		} else {
			step = 1;
		}
	}

	bsp_uDelay(SLED_CYCLE_US/(SLED_TEST_COUNT_MAX+1));

	return true;
}

bool apb3_LvdsLoopbackTest(test_state_t test_state)
{
	static bool lvdsError = false;
	u32 status;
	u32 status_change;

	if (test_state == test_state_end) {
		apb3_clearControl(APB0, CONT_LVDS_TXPLL_RST_N | CONT_LVDS_RXPLL_RST_N |
								CONT_LVDS_TEST1 | CONT_LVDS_TEST2 | CONT_LVDS_TEST3 | CONT_LVDS_TEST4 | CONT_LVDS_TEST5);
		SLED_BLK(2, 1);
		lvdsError = false;
		return false;
	}

	if (lvdsError) {
		return true;
	}

	status_change = apb3_readStatusChange(APB0);
	status = apb3_readStatus(APB0);

	if (test_state == test_state_start) {
		SLED_YEL(2, 1);
		bsp_uDelay(100000);
		apb3_clearStatusChange(APB0, STAT_LVDS_TXPLL_LOCKED | STAT_LVDS_RXPLL_LOCKED | STAT_LVDS_TEST_RX | STAT_LVDS_TEST_ERR);
		apb3_setControl(APB0, CONT_LVDS_TXPLL_RST_N);
	}

	if ((status_change & STAT_LVDS_TEST_ERR) && (status & STAT_LVDS_TEST_ERR)) {
		SLED_RED(2, 1);
		apb3_clearStatusChange(APB0, STAT_LVDS_TEST_ERR);
		bsp_printf("RX Data error\n");
		lvdsError = true;
		return true;
	}

	if ((status_change & STAT_LVDS_TXPLL_LOCKED) && (status & STAT_LVDS_TXPLL_LOCKED)) {
		bsp_printf("TX PLL locked\n");
		apb3_clearStatusChange(APB0, STAT_LVDS_TXPLL_LOCKED);
		apb3_setControl(APB0, CONT_LVDS_RXPLL_RST_N);
	}

	if ((status_change & STAT_LVDS_RXPLL_LOCKED) && (status & STAT_LVDS_RXPLL_LOCKED)) {
		bsp_printf("RX PLL locked\n");
		apb3_clearStatusChange(APB0, STAT_LVDS_RXPLL_LOCKED);
		apb3_setControl(APB0, CONT_LVDS_TEST1 | CONT_LVDS_TEST2 | CONT_LVDS_TEST3 | CONT_LVDS_TEST4 | CONT_LVDS_TEST5);
	}

	if ((status_change & STAT_LVDS_TEST_RX) && (status & STAT_LVDS_TEST_RX) && (!(status_change & STAT_LVDS_TEST_ERR))) {
		SLED_GRN(2, 1);
		apb3_clearStatusChange(APB0, STAT_LVDS_TEST_RX);
		bsp_printf("RX Data receiving\n");
	}

	return true;
}
