// The MIT License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software
// and associated documentation files (the “Software”), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge, publish, distribute,
// sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
// NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
#include <stdbool.h>
#include <stdint.h>
#include <math.h>

#include <bsp.h>
#include <spi.h>

#include "adda.h"
#include "apb3.h"
#include "gpio.h"

#define ADDA_SPI SYSTEM_SPI_1_IO_CTRL

uint32_t mcp3008_read(mcp3008_t s)
{
#if DUMMY_READ	// Dummy Read (Select internal multiplexor)
	spi_select(ADDA_SPI, 1);
	spi_write(ADDA_SPI, s.arr[2]);
	spi_write(ADDA_SPI, s.arr[1]);
	spi_write(ADDA_SPI, s.arr[0]);
	spi_diselect(ADDA_SPI, 1);
#endif
	spi_select(ADDA_SPI, 1);
	spi_write(ADDA_SPI, s.arr[2]);
	s.arr[1] = spi_writeRead(ADDA_SPI, s.arr[1]);
	s.arr[0] = spi_writeRead(ADDA_SPI, s.arr[0]);
	spi_diselect(ADDA_SPI, 1);

	return (uint32_t)s.data;
}

#if FACTORY == true// Execute only during factory inspection.
const mcp48feb_t MCP48FEB_FACT_SET_TB [] = {
	{.ad = MCP48_NV_DAC0_ADDR, .c = MCP48_CMD_WRITE, .cmderr = MCP48_CMD_VALID, .dac.b15_10 = 0, .dac.d = 0},
#if DA_CHANNELS == 2
	{.ad = MCP48_NV_DAC1_ADDR, .c = MCP48_CMD_WRITE, .cmderr = MCP48_CMD_VALID, .dac.b15_10 = 0, .dac.d = 0},
#endif
	{.ad = MCP48_NV_VREF_ADDR, .c = MCP48_CMD_WRITE, .cmderr = MCP48_CMD_VALID, .vr.b15_8 = 0, .vr.b7_2 = 0, .vr.vr0 = VDD_UNBUFFERED},
	{.ad = MCP48_NV_POW_DOWN_ADDR, .c = MCP48_CMD_WRITE, .cmderr = MCP48_CMD_VALID, .pd.b15_8 = 0, .pd.b7_2 = 0, .pd.pd0 = POWER_DOWN_VOUT_OPEN},
	{.ad = MCP48_NV_GAIN_ADDR, .c = MCP48_CMD_WRITE, .cmderr = MCP48_CMD_VALID, .g.b15_9 = 0, .g.g0 = 0, .g.b7_0 = 0}
};
#endif

mcp48feb_t mcp48_write_read(mcp48feb_t s)
{
	spi_select(ADDA_SPI, 2);
	s.arr[2] = spi_writeRead(ADDA_SPI, s.arr[2]);
	s.arr[1] = spi_writeRead(ADDA_SPI, s.arr[1]);
	s.arr[0] = spi_writeRead(ADDA_SPI, s.arr[0]);
	spi_diselect(ADDA_SPI, 2);

	return s;
}

mcp48feb_t mcp48_nvm_write_read(mcp48feb_t s)
{
	while (MCP48FEB_GET_STAT().gs.eewa);
	return mcp48_write_read(s);
}

#if FACTORY // Execute only during factory inspection.
void mcp48feb_factory_set(void)
{
	for (int i = 0; i < sizeof(MCP48FEB_FACT_SET_TB)/sizeof(mcp48feb_t); i++) {
		mcp48_nvm_write_read(MCP48FEB_FACT_SET_TB[i]);
	}
}
#endif

// Initial value of APB3 register.
// Initial value of MCP48FEB non-volatile memory.
static bool ext_vref_set = false;

void adda_ext_vref_sel(bool ext_vref)
{
	if (ext_vref) {
		if (!ext_vref_set) {
			// DAC EXT3V -> ADC EXT3V
			MCP48FEB_SET_VREF(VREF_BUFFERED);
			apb3_clearControl(APB0, CONT_AREF_EXT_SEL_N);
		}
	} else {
		if (!ext_vref_set) {
			// ADC A3R3V -> DAC A3R3V
			apb3_setControl(APB0, CONT_AREF_EXT_SEL_N);
			MCP48FEB_SET_VREF(VDD_UNBUFFERED);
		}
	}
	ext_vref_set = ext_vref;
}

uint32_t calc_mon_current_ma(uint32_t ad_val, uint32_t adc_res)
{
	uint32_t vref_mv;
	uint32_t rclim;
	uint32_t iset_ma;
	u32 status;

	if (ext_vref_set) {
		vref_mv = 3000;
	} else {
		vref_mv = 3300;
	}

	status = apb3_readStatus(APB0);
	if (status & STAT_VIN_MON_N) {
		iset_ma = 500;
		if (!(status & CONT_ISET_1R0A_N)) {
			iset_ma += 1000;
		}
		if (!(status & CONT_ISET_1R5A_N)) {
			iset_ma += 1500;
		}
		switch (iset_ma) {
		case 1500:
			rclim = (uint32_t)roundf(RCLIM_1500MA);
			break;
		case 2000:
			rclim = (uint32_t)roundf(RCLIM_2000MA);
			break;
		case 3000:
			rclim = (uint32_t)roundf(RCLIM_3000MA);
			break;
		default:
			rclim = (uint32_t)roundf(RCLIM_500MA);
			break;
		}
	} else {
		rclim = (uint32_t)roundf(RCLIM_VIN);
	}

	return vref_mv * ad_val / adc_res * 1000000 / (uint32_t)roundf(CMON_GAIN * 1000) / rclim;
}
