// The MIT License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software
// and associated documentation files (the “Software”), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge, publish, distribute,
// sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
// NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
#include <stdbool.h>

#include <bsp.h>
#include <spi.h>

#include "flash.h"

static bool FlashWaitBusy(int timeout_ms)
{
	bool retval = false;

    while(timeout_ms) {
        bsp_uDelay(1000);
        if (FlashReadStatusReg1().busy) {
        	timeout_ms--;
        } else {
        	retval = true;
            break;
        }
    }
    if (!retval) {
        bsp_printf("Flash Time out \r\n");
    }
    return retval;
}

void FlashWriteEnable(void)
{
    spi_select(FL_SPI, 0);
    spi_write(FL_SPI, FLASH_WRITE_ENABLE);
    spi_diselect(FL_SPI, 0);
}

void FlashReleasePowerDown(void)
{
    spi_select(FL_SPI, 0);
    spi_write(FL_SPI, FLASH_RELEASE_POWER_DOWN);
    spi_diselect(FL_SPI, 0);
    bsp_uDelay(FLASH_POWER_DOWN_TO_STAND_BY_US);
}

void FlashReadData(FLASH_t arg)
{
	spi_select(FL_SPI, 0);
#if FAST_READ
	spi_write(FL_SPI, FLASH_FAST_READ);
#else
	spi_write(FL_SPI, FLASH_READ_DATA);
#endif
	spi_write(FL_SPI, arg.addr.u8[2]);
	spi_write(FL_SPI, arg.addr.u8[1]);
	spi_write(FL_SPI, arg.addr.u8[0]);
#if FAST_READ
	spi_write(FL_SPI, 0);
#endif
	for (int i = 0; i < arg.len; i++) {
    	arg.data[i] = spi_read(FL_SPI);
	}
	spi_diselect(FL_SPI, 0);
}

bool FlashPageProgram(FLASH_t arg)
{
	FlashWriteEnable();
	spi_select(FL_SPI, 0);
	spi_write(FL_SPI, FLASH_PAGE_PROGRAM);
	spi_write(FL_SPI, arg.addr.u8[2]);
	spi_write(FL_SPI, arg.addr.u8[1]);
	spi_write(FL_SPI, arg.addr.u8[0]);
    for (int i = 0; i < arg.len; i++) {
    	spi_write(FL_SPI, arg.data[i]);
	}
	spi_diselect(FL_SPI, 0);

	return FlashWaitBusy(FLASH_WRITE_TIMEOUT_MS);
}

bool FlashSectorErase(FLASH_ADDR_t Addr)
{
	FlashWriteEnable();
    spi_select(FL_SPI, 0);
    spi_write(FL_SPI, FLASH_SECTOR_ERASE);
    spi_write(FL_SPI, Addr.u8[2]);
    spi_write(FL_SPI, Addr.u8[1]);
    spi_write(FL_SPI, Addr.u8[0]);
    spi_diselect(FL_SPI, 0);
    return FlashWaitBusy(FLASH_SECTOR_ERASE_TIMEOUT_MS);
}

bool FlashBlockErase(FLASH_ADDR_t Addr)
{
	FlashWriteEnable();
    spi_select(FL_SPI, 0);
    spi_write(FL_SPI, FLASH_BLOCK_ERASE);
    spi_write(FL_SPI, Addr.u8[2]);
    spi_write(FL_SPI, Addr.u8[1]);
    spi_write(FL_SPI, Addr.u8[0]);
    spi_diselect(FL_SPI, 0);
    return FlashWaitBusy(FLASH_BLOCK_ERASE_TIMEOUT_MS);
}

FLASH_ST1_t FlashReadStatusReg1(void)
{
	FLASH_ST1_t status;

    spi_select(FL_SPI, 0);
    spi_write(FL_SPI, FLASH_READ_STATUS_REG1);
    status.u8 = spi_read(FL_SPI);
    spi_diselect(FL_SPI, 0);

    return status;
}

void FlashPowerDown(void)
{
    spi_select(FL_SPI, 0);
    //Power down
    spi_write(FL_SPI, FLASH_POWER_DOWN);
    spi_diselect(FL_SPI, 0);
}
