// The MIT License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software
// and associated documentation files (the “Software”), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge, publish, distribute,
// sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
// NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
#include <stdbool.h>

#include <bsp.h>
#include <plic.h>
#include <clint.h>
#include <riscv.h>

#include "timer.h"
#include "power.h"
#include "lib/apb3.h"
#include "lib/gpio.h"

#define POWER_OFF_LONG_PRESS_TIME		2.0f		// 2sec

static void powerOffProc(void)
{
	SLED_BLK(0, 0);
	SLED_BLK(1, 0);
	SLED_BLK(2, 1);

	apb3_clearControl(APB0, CONT_PWR_LED_R_N);
	apb3_setControl(APB0, CONT_PWR_LED_G_N);
	apb3_setControl(APB0, CONT_PWR_LED_B_N);

	apb3_clearControl(APB0, CONT_POW_CNT_BIT);

	while (1);
}

static bool readDipsw(u32* dipsw)
{
	bool retval = false;
	u32 dipsw_new;
	static u32 dipsw_buf = ~0;

	dipsw_new = gpio32_getInput(DIPSW1_BIT | DIPSW2_BIT | DIPSW3_BIT | DIPSW4_BIT);
	if (dipsw_new == dipsw_buf) {
		if (*dipsw != dipsw_new) {
			*dipsw = dipsw_new;
			retval = true;
		}
	}
	dipsw_buf = dipsw_new;
	return retval;
}

static void sledCurrentLimit(u32 limCurrent)
{
	switch (limCurrent) {
	case 500:
		SLED_RED(0, 1);
		break;
	case 1500:
		SLED_GRN(0, 1);
		break;
	case 2000:
		SLED_BLU(0, 1);
		break;
	case 3000:
		SLED_CYN(0, 1);
		break;
	default :
		SLED_BLK(0, 1);
		break;
	}
}

static void powerSourceDetection(u32 status)
{
	int limCurrent;

	if (status & STAT_TYPEC_TCC1) {
		if (status & STAT_TYPEC_TCC0) {
			limCurrent = 3000;
			bsp_printf("USB Type-C 3.0A");
		} else {
			limCurrent = 1500;
			bsp_printf("USB Type-C 1.5A");
		}
	} else if ((status & STAT_USB_CHR0) && (status & STAT_USB_CHR1) && (status & STAT_USB_CHR_EN)) {
		limCurrent = 1500;
		bsp_printf("USB BC1.2 1.5A");
	} else {
		limCurrent = 500;
		bsp_printf("USB Normal");
	}

	switch (gpio32_getInput(DIPSW1_BIT | DIPSW2_BIT)) {
	case DIPSW1_BIT | DIPSW2_BIT:
		break;
	case DIPSW2_BIT:
		if (limCurrent > 2000) {
			limCurrent = 2000;
		}
		break;
	case DIPSW1_BIT:
		if (limCurrent > 1500) {
			limCurrent = 1500;
		}
		break;
	default:
		limCurrent = 500;
		break;
	}

	switch (limCurrent) {
	case 3000:
		apb3_clearControl(APB0, CONT_ISET_1R5A_N | CONT_ISET_1R0A_N);
		break;
	case 2000:
		apb3_clearControl(APB0, CONT_ISET_1R5A_N);
		apb3_setControl(APB0, CONT_ISET_1R0A_N);
		break;
	case 1500:
		apb3_clearControl(APB0, CONT_ISET_1R0A_N);
		apb3_setControl(APB0, CONT_ISET_1R5A_N);
		break;
	default:
		apb3_setControl(APB0, CONT_ISET_1R5A_N | CONT_ISET_1R0A_N);
		break;
	}

	if (status & STAT_VBUS_MON_N) {
		limCurrent = 3000;
		bsp_printf(" / VIN : ILim=%dmA\n", limCurrent);
	} else {
		bsp_printf(" / VBUS : ILim=%dmA\n", limCurrent);
	}

	sledCurrentLimit(limCurrent);
}

static void powerProc(bool init)
{
	u32 status;
	u32 status_change;
	static int powerOffSwCount = -1;
	static u32 dipsw = ~0;

	status_change = apb3_readStatusChange(APB0);
	apb3_clearStatusChange(APB0, STAT_VIN_MON_N | STAT_TYPEC_SWMONI);
	status = apb3_readStatus(APB0);

	if (status & STAT_PWR_SW_N_BIT) {
		powerOffSwCount = 0;
	} else {
		if (powerOffSwCount >= 0) {
			powerOffSwCount++;
		}
		if (powerOffSwCount >= (int)(SYSTEM_CLINT_HZ * POWER_OFF_LONG_PRESS_TIME / TIMER_TICK_DELAY)) {
			powerOffProc();
		}
	}

	if (status_change & STAT_VIN_MON_N) {
		if (status & STAT_VIN_MON_N) {
			apb3_setControl(APB0, CONT_PWR_LED_G_N);
		} else {
			apb3_clearControl(APB0, CONT_PWR_LED_G_N);
			apb3_setControl(APB0, CONT_PWR_LED_R_N);
		    bsp_printf("VIN Power detect\n");
		}
	}

	if (status_change & STAT_TYPEC_SWMONI) {
		if (status & STAT_TYPEC_SWMONI) {
			apb3_clearControl(APB0, CONT_PWR_LED_B_N);
			apb3_setControl(APB0, CONT_PWR_LED_R_N);
		    bsp_printf("VBUS Power detect\n");
		} else {
			apb3_setControl(APB0, CONT_PWR_LED_B_N);
		}
	}

	if ((init) || (readDipsw(&dipsw))) {
		powerSourceDetection(status);
		if (dipsw & DIPSW3_BIT) {
			if (!(status & CONT_USB_PWR_DIS)) {
				apb3_setControl(APB0, CONT_USB_PWR_DIS);
				bsp_printf("USB Power disable\n");
			}
		} else {
			if (status & CONT_USB_PWR_DIS) {
				apb3_clearControl(APB0, CONT_USB_PWR_DIS);
				bsp_printf("USB Power enable\n");
			}
		}

		if (dipsw & DIPSW4_BIT) {
			if (status & CONT_AREF_EXT_SEL_N) {
				apb3_clearControl(APB0, CONT_AREF_EXT_SEL_N);
				bsp_printf("External AREF\n");
			}
		} else {
			if (!(status & CONT_AREF_EXT_SEL_N)) {
				apb3_setControl(APB0, CONT_AREF_EXT_SEL_N);
				bsp_printf("Internal AREF\n");
			}
		}
	}
}

static void powerTimerProc(void)
{
	powerProc(false);
}

void powerInterruptProc(void)
{
	apb3_clearInterrupt(APB0, STAT_VBUS_MON_N | STAT_TYPEC_TCC0 | STAT_TYPEC_TCC1 | STAT_USB_CHR0 | STAT_USB_CHR1 | STAT_USB_CHR_EN);
	powerSourceDetection(apb3_readStatus(APB0));
}

void powerInit(void)
{
	apb3_setInterruptUpedge(APB0, STAT_VBUS_MON_N | STAT_TYPEC_TCC0 | STAT_TYPEC_TCC1 | STAT_USB_CHR0 | STAT_USB_CHR1 | STAT_USB_CHR_EN);
	apb3_setInterruptDownedge(APB0, STAT_VBUS_MON_N | STAT_TYPEC_TCC0 | STAT_TYPEC_TCC1 | STAT_USB_CHR0 | STAT_USB_CHR1 | STAT_USB_CHR_EN);
	apb3_clearInterrupt(APB0, STAT_VBUS_MON_N | STAT_TYPEC_TCC0 | STAT_TYPEC_TCC1 | STAT_USB_CHR0 | STAT_USB_CHR1 | STAT_USB_CHR_EN);

	plic_set_enable(BSP_PLIC, BSP_PLIC_CPU_0, SYSTEM_PLIC_USER_INTERRUPT_A_INTERRUPT, 1);
    plic_set_priority(BSP_PLIC, SYSTEM_PLIC_USER_INTERRUPT_A_INTERRUPT, 1);

    powerProc(true);

	machineTimerSetInterruptProc(powerTimerProc);
}
