// The MIT License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software
// and associated documentation files (the “Software”), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge, publish, distribute,
// sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
// NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
#include <stdbool.h>
#include <stdint.h>
#include <math.h>

#include "bsp.h"
#include "clint.h"
#include "spi.h"
#include "test.h"
#include "timer.h"
#include "lib/spi.h"
#include "lib/apb3.h"
#include "lib/adda.h"
#include "lib/flash.h"

#define AD_RESOLUTION				1024

#define DA_SIN_STEPS				256
#define ADDA_SAMPLING_PERIOD_MS		(1000 / DA_SIN_STEPS)	// 1000ms -> 1Hz

#define DA_RESOLUTION				1024
#define DA_SIN_LEVEL				1000

#define CURR_MON_OVER_SAMPLING		64		// Over sampling order

#define SIN_TBL_USE					true

#if SIN_TBL_USE == true
#define DASIN(s)		(u16)(sin(2.0f*M_PI*s/DA_SIN_STEPS) * (DA_SIN_LEVEL/2) + (DA_RESOLUTION/2))
const u16 SIN_TBL[DA_SIN_STEPS] = {
	DASIN(0),   DASIN(1),   DASIN(2),   DASIN(3),   DASIN(4),   DASIN(5),   DASIN(6),   DASIN(7),   DASIN(8),   DASIN(9),
	DASIN(10),  DASIN(11),  DASIN(12),  DASIN(13),  DASIN(14),  DASIN(15),  DASIN(16),  DASIN(17),  DASIN(18),  DASIN(19),
	DASIN(20),  DASIN(21),  DASIN(22),  DASIN(23),  DASIN(24),  DASIN(25),  DASIN(26),  DASIN(27),  DASIN(28),  DASIN(29),
	DASIN(30),  DASIN(31),  DASIN(32),  DASIN(33),  DASIN(34),  DASIN(35),  DASIN(36),  DASIN(37),  DASIN(38),  DASIN(39),
	DASIN(40),  DASIN(41),  DASIN(42),  DASIN(43),  DASIN(44),  DASIN(45),  DASIN(46),  DASIN(47),  DASIN(48),  DASIN(49),
	DASIN(50),  DASIN(51),  DASIN(52),  DASIN(53),  DASIN(54),  DASIN(55),  DASIN(56),  DASIN(57),  DASIN(58),  DASIN(59),
	DASIN(60),  DASIN(61),  DASIN(62),  DASIN(63),  DASIN(64),  DASIN(65),  DASIN(66),  DASIN(67),  DASIN(68),  DASIN(69),
	DASIN(70),  DASIN(71),  DASIN(72),  DASIN(73),  DASIN(74),  DASIN(75),  DASIN(76),  DASIN(77),  DASIN(78),  DASIN(79),
	DASIN(80),  DASIN(81),  DASIN(82),  DASIN(83),  DASIN(84),  DASIN(85),  DASIN(86),  DASIN(87),  DASIN(88),  DASIN(89),
	DASIN(90),  DASIN(91),  DASIN(92),  DASIN(93),  DASIN(94),  DASIN(95),  DASIN(96),  DASIN(97),  DASIN(98),  DASIN(99),
	DASIN(100), DASIN(101), DASIN(102), DASIN(103), DASIN(104), DASIN(105), DASIN(106), DASIN(107), DASIN(108), DASIN(109),
	DASIN(110), DASIN(111), DASIN(112), DASIN(113), DASIN(114), DASIN(115), DASIN(116), DASIN(117), DASIN(118), DASIN(119),
	DASIN(120), DASIN(121), DASIN(122), DASIN(123), DASIN(124), DASIN(125), DASIN(126), DASIN(127), DASIN(128), DASIN(129),
	DASIN(130), DASIN(131), DASIN(132), DASIN(133), DASIN(134), DASIN(135), DASIN(136), DASIN(137), DASIN(138), DASIN(139),
	DASIN(140), DASIN(141), DASIN(142), DASIN(143), DASIN(144), DASIN(145), DASIN(146), DASIN(147), DASIN(148), DASIN(149),
	DASIN(150), DASIN(151), DASIN(152), DASIN(153), DASIN(154), DASIN(155), DASIN(156), DASIN(157), DASIN(158), DASIN(159),
	DASIN(160), DASIN(161), DASIN(162), DASIN(163), DASIN(164), DASIN(165), DASIN(166), DASIN(167), DASIN(168), DASIN(169),
	DASIN(170), DASIN(171), DASIN(172), DASIN(173), DASIN(174), DASIN(175), DASIN(176), DASIN(177), DASIN(178), DASIN(179),
	DASIN(180), DASIN(181), DASIN(182), DASIN(183), DASIN(184), DASIN(185), DASIN(186), DASIN(187), DASIN(188), DASIN(189),
	DASIN(190), DASIN(191), DASIN(192), DASIN(193), DASIN(194), DASIN(195), DASIN(196), DASIN(197), DASIN(198), DASIN(199),
	DASIN(200), DASIN(201), DASIN(202), DASIN(203), DASIN(204), DASIN(205), DASIN(206), DASIN(207), DASIN(208), DASIN(209),
	DASIN(210), DASIN(211), DASIN(212), DASIN(213), DASIN(214), DASIN(215), DASIN(216), DASIN(217), DASIN(218), DASIN(219),
	DASIN(220), DASIN(221), DASIN(222), DASIN(223), DASIN(224), DASIN(225), DASIN(226), DASIN(227), DASIN(228), DASIN(229),
	DASIN(230), DASIN(231), DASIN(232), DASIN(233), DASIN(234), DASIN(235), DASIN(236), DASIN(237), DASIN(238), DASIN(239),
	DASIN(240), DASIN(241), DASIN(242), DASIN(243), DASIN(244), DASIN(245), DASIN(246), DASIN(247), DASIN(248), DASIN(249),
	DASIN(250), DASIN(251), DASIN(252), DASIN(253), DASIN(254), DASIN(255)
};
#endif

bool spi_adcDacTest(test_state_t test_state)
{
	if (test_state == test_state_end) {
		return false;
	}

	for (int j = 0; j < DA_SIN_STEPS; j++) {
		int dac_val;
#if SIN_TBL_USE == true
		dac_val = SIN_TBL[j];
#else
		float rad;
		rad = 2.0f * M_PI * ((float)j / DA_SIN_STEPS);
		dac_val = sinf(rad) * (DA_SIN_LEVEL/2) + (DA_RESOLUTION/2);
#endif
		MCP48FEB_SET_DAC1(dac_val);
#if DA_CHANNELS == 2
#if SIN_TBL_USE == true
		dac_val = SIN_TBL[(j + DA_SIN_STEPS/2) % DA_SIN_STEPS];
#else
		rad = 2.0f * M_PI * ((float)(j+DA_SIN_STEPS/2) / DA_SIN_STEPS);
		dac_val = sinf(rad) * (DA_SIN_LEVEL/2) + (DA_RESOLUTION/2);
#endif
		MCP48FEB_SET_DAC2(dac_val);
#endif
		bsp_printf("ADC Read :");
		for (int k = 0; k < AD_CHANNELS; k++) {
			bsp_printf(" %d", MCP3008_READ(k));
		}
		bsp_printf("\r\n");
		bsp_uDelay(ADDA_SAMPLING_PERIOD_MS * 1000);
	}
	return true;
}

bool spi_currentMonitorTest(test_state_t test_state)
{
	uint32_t val = 0;
	uint32_t curr;

	if (test_state == test_state_end) {
		blinkLed(0);
		return false;
	}

	if (test_state == test_state_start) {
		blinkLed(-1);	// Stop blinking LED
	}

	for (int i = 0; i < CURR_MON_OVER_SAMPLING; i++) {
		bsp_uDelay(ADDA_SAMPLING_PERIOD_MS * 1000);
		if (apb3_readStatus(APB0) & STAT_VBUS_MON_N) {
			val += MCP3008_READ(AD_VIN_CURR);
		} else {
			val += MCP3008_READ(AD_VBUS_CURR);
		}
	}
	curr = calc_mon_current_ma(val, AD_RESOLUTION * CURR_MON_OVER_SAMPLING);
	bsp_printf("Current:%ddec/%dmA\r\n", (int)(val / CURR_MON_OVER_SAMPLING), (int)curr);

	return true;
}

static void putHexChar(u8 c)
{
	static const char hex_tbl[] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};

	bsp_putChar(hex_tbl[c/16]);
	bsp_putChar(hex_tbl[c%16]);
}

static void flashPageProgramTest(FLASH_ADDR_t addr, u8 firstData)
{
	u8 data[FLASH_PAGE_SIZE];

	for (int i = 0; i < FLASH_PAGE_SIZE; i++) {
		data[i] = firstData + i;
	}

	FlashPageProgram((FLASH_t){.addr = addr, .data = data, .len = FLASH_PAGE_SIZE});

	bsp_printf("Page Program 0x%08x : ", addr.u32);

	for (int i = 0; i < 4; i++) {
		putHexChar(data[i]);
		bsp_printf(", ");
	}
	bsp_printf(",,, ");
	putHexChar(data[FLASH_PAGE_SIZE-1]);
	bsp_putChar('\n');
}

static void flashReadDataTest(FLASH_ADDR_t addr, size_t len)
{
	u8 data[len];

	FlashReadData((FLASH_t){.addr = addr, .data = data, .len = len});
	bsp_printf("Read Data 0x%08x\n", addr.u32);

	for (int i = 0; i < len; i++) {
		if (i % 16 == 0) {
			bsp_printf("0x%08x : ", addr.u32 + i);
		}
		putHexChar(data[i]);
		if (i % 16 == 15) {
			bsp_putChar('\n');
		} else {
			bsp_printf(", ");
		}
	}
	bsp_putChar('\n');
}

bool spi_flashAccessTest(test_state_t test_state)
{
	if (test_state == test_state_start) {

		FlashReleasePowerDown();
		bsp_printf("Release Power Down\n");

		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 0)}, FLASH_PAGE_SIZE*2);
		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 1, 0)}, FLASH_PAGE_SIZE);
		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, FLASH_SECT_NUMS-1, FLASH_PAGE_NUMS-1)}, FLASH_PAGE_SIZE);

		FlashBlockErase((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 0)});
		bsp_printf("Block Erase %08x\n", (FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 0)}.u32);

		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 0)}, FLASH_PAGE_SIZE*2);
		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 1, 0)}, FLASH_PAGE_SIZE);
		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, FLASH_SECT_NUMS-1, FLASH_PAGE_NUMS-1)}, FLASH_PAGE_SIZE);

		flashPageProgramTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 0)}, (u8)clint_getTime(BSP_CLINT));
		flashPageProgramTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 1)}, (u8)clint_getTime(BSP_CLINT));
		flashPageProgramTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 1, 0)}, (u8)clint_getTime(BSP_CLINT));
		flashPageProgramTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, FLASH_SECT_NUMS-1, FLASH_PAGE_NUMS-1)}, (u8)clint_getTime(BSP_CLINT));

		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 0)}, FLASH_PAGE_SIZE*2);
		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 1, 0)}, FLASH_PAGE_SIZE);
		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, FLASH_SECT_NUMS-1, FLASH_PAGE_NUMS-1)}, FLASH_PAGE_SIZE);

		FlashSectorErase((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 0)});
		bsp_printf("Sector Erase %08x\n", (FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 0)}.u32);

		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 0)}, FLASH_PAGE_SIZE*2);
		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 1, 0)}, FLASH_PAGE_SIZE);
		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, FLASH_SECT_NUMS-1, FLASH_PAGE_NUMS-1)}, FLASH_PAGE_SIZE);

	}

	if (test_state == test_state_end) {
		FlashBlockErase((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 0)});
		bsp_printf("Block Erase %08x\n", (FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 0)}.u32);

		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 0, 0)}, FLASH_PAGE_SIZE*2);
		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, 1, 0)}, FLASH_PAGE_SIZE);
		flashReadDataTest((FLASH_ADDR_t){.p = FLASH_ADDR(FLASH_BLOCK_NUMS-1, FLASH_SECT_NUMS-1, FLASH_PAGE_NUMS-1)}, FLASH_PAGE_SIZE);

		FlashPowerDown();
		bsp_printf("Power Down\n");
		return false;
	}

	return true;
}

void spiTestInit(void)
{
	spiInit();

#if FACTORY // Execute only during factory inspection.
	mcp48feb_factory_set();
#endif

	MCP48FEB_SET_DAC1(DA_RESOLUTION/2);
#if DA_CHANNELS == 2
	MCP48FEB_SET_DAC2(DA_RESOLUTION/2);
#endif

	MCP48FEB_SET_PD(NORMAL_OPERATION);
}

