// The MIT License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software
// and associated documentation files (the “Software”), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge, publish, distribute,
// sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
// NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
#include <stdbool.h>
#include <math.h>

#include <bsp.h>

#include "timer.h"
#include "lib/gpio.h"

static volatile uint64_t timerCmp;

#define LED_BLINK_TIME	1.0f	// 1sec

#define INT_PROC_NUMS	10

void (*intProc[INT_PROC_NUMS])(void);

void blinkLed(int set)
{
#ifdef LED_BLINK_TIME
	static int blinkLedCount = 0;

	if (set < 0) {
		blinkLedCount = -1;
		gpio32_clearOutput(LED1_BIT);
	} else if (set == 0) {
		if (blinkLedCount < 0) {
			blinkLedCount = 0;
		}
	} else {
		if (blinkLedCount == 0) {
			gpio32_setOutput(LED1_BIT);
		} else if (blinkLedCount == (int)roundf(SYSTEM_CLINT_HZ * LED_BLINK_TIME / TIMER_TICK_DELAY / 2)) {
			gpio32_clearOutput(LED1_BIT);
		}
		if (blinkLedCount >= 0) {
			blinkLedCount++;
			if (blinkLedCount == (int)roundf(SYSTEM_CLINT_HZ * LED_BLINK_TIME / TIMER_TICK_DELAY)) {
				blinkLedCount = 0;
			}
		}
	}
#endif
}

static void scheduleTimer(void)
{
    timerCmp += TIMER_TICK_DELAY;
    clint_setCmp(BSP_CLINT, timerCmp, 0);
}

void machineTimerInterrupt(void)
{
	scheduleTimer();
	blinkLed(1);

	for (int i = 0; i < INT_PROC_NUMS; i++)
	{
		if (intProc[i] != NULL) {
			intProc[i]();
		}
	}
}

int machineTimerSetInterruptProc(void(*proc)(void))
{
	for (int i = 0; i < INT_PROC_NUMS; i++)
	{
		if (intProc[i] == NULL) {
			intProc[i] = proc;
			return i;
		}
	}
	return -1;
}

void machineTimerInit(void)
{
    timerCmp = clint_getTime(BSP_CLINT);
    scheduleTimer();
	blinkLed(0);
}

void startMachineTimer(u32* tim)
{
	*tim = clint_getTimeLow(BSP_CLINT);
}

s32 checkMachineTimer(u32* tim, u32 usec)
{
	s32 ret;
    u32 mTimePerUsec = BSP_CLINT_HZ/1000000;
	u32 limit = *tim + usec * mTimePerUsec;
	ret = limit - clint_getTimeLow(BSP_CLINT);
	if (ret <= 0) {
		*tim = limit;
	}
	return ret;
}



