// The MIT License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software
// and associated documentation files (the “Software”), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge, publish, distribute,
// sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
// NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
#include <stdbool.h>

#include <bsp.h>
#include <plic.h>
#include <clint.h>
#include <riscv.h>

#include "uart.h"
#include "timer.h"
#include "lib/apb3.h"
#include "lib/gpio.h"

#define UART_LED				true
#define UART_FLOW				true
#define UART_RTS_ON_OCCUPANCY	16

#define UART_A_SAMPLE_PER_BAUD 	8
#define CORE_HZ 				BSP_CLINT_HZ

static void (*uartReceiveProc)(char c) = NULL;

#if UART_LED == true
#define SET_UART_LED(led)		gpio32_setOutput(led)
#define CLR_UART_LED(led)		gpio32_clearOutput(led)
#else
#define SET_UART_LED(led)
#define CLR_UART_LED(led)
#endif

void uartSetReceiveProc(void(*proc)(char))
{
	uartReceiveProc = proc;
}

void uartSendChar(char c)
{
	uart_write(BSP_UART_TERMINAL, c);
}

void uartInterrupt(void)
{
	char c;

    if (uart_status_read(BSP_UART_TERMINAL) & BIT_9){
        uart_status_write(BSP_UART_TERMINAL, uart_status_read(BSP_UART_TERMINAL) & ~BIT_1);
#if UART_FLOW == true
        if (!(apb3_readStatus(APB0) & CONT_USB_VCP_RTS)) {
			if (uart_readOccupancy(BSP_UART_TERMINAL) >= UART_RTS_ON_OCCUPANCY) {
				apb3_setControl(APB0, CONT_USB_VCP_RTS);
#if UART_LED == true
				gpio32_clearOutput(LED2_BIT);
#endif
			}
		}
#endif
		c = uart_read(BSP_UART_TERMINAL);

		if (uartReceiveProc != NULL) {
			uartReceiveProc(c);
		}
        uart_status_write(BSP_UART_TERMINAL,uart_status_read(BSP_UART_TERMINAL) | BIT_1);
    }
}

static void uartFlowControl(void)
{
	u32 status_change;
	u32 status;

	status_change = apb3_readStatusChange(APB0);
	apb3_clearStatusChange(APB0, CONT_USB_VCP_RTS | STAT_USB_VCP_CTS | STAT_USB_SUSPENDED);
	status = apb3_readStatus(APB0);

#if UART_FLOW == true
	if (status & CONT_USB_VCP_RTS) {
		if (uart_readOccupancy(BSP_UART_TERMINAL) == 0) {
			apb3_clearControl(APB0, CONT_USB_VCP_RTS);
#if UART_LED == true
			gpio32_setOutput(LED2_BIT);
#endif
    	}
	}
#endif

	if (status_change & STAT_USB_VCP_CTS) {
		if (status & STAT_USB_VCP_CTS) {
#if UART_LED == true
			gpio32_clearOutput(LED3_BIT);
#endif
		} else {
#if UART_LED == true
			gpio32_setOutput(LED3_BIT);
#endif
			uartReceiveProc('0');
		}
	}

#if UART_LED == true
	if (status_change & STAT_USB_SUSPENDED) {
		if (status & STAT_USB_SUSPENDED) {
			gpio32_clearOutput(LED4_BIT);
		} else {
			gpio32_setOutput(LED4_BIT);
		}
	}
#endif
}

void uartInit(void)
{
    uart_applyConfig(BSP_UART_TERMINAL,
    	&(Uart_Config) {
			.dataLength = BITS_8,
			.parity = NONE,
			.stop = ONE,
			.clockDivider = CORE_HZ / (115200 * UART_A_SAMPLE_PER_BAUD) - 1
    	}
    );

    uart_RX_NotemptyInterruptEna(BSP_UART_TERMINAL, 1);

    plic_set_enable(BSP_PLIC, BSP_PLIC_CPU_0, SYSTEM_PLIC_SYSTEM_UART_0_IO_INTERRUPT, 1);
    plic_set_priority(BSP_PLIC, SYSTEM_PLIC_SYSTEM_UART_0_IO_INTERRUPT, 1);

	apb3_clearControl(APB0, CONT_USB_VCP_RTS);
	machineTimerSetInterruptProc(uartFlowControl);

#if UART_LED == true
	gpio32_setOutput(LED2_BIT);
#endif
}
