// The CERN-OHL-P v2 License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// This source describes Open Hardware and is licensed under the CERN-OHL-P v2.
// You may redistribute and modify this documentation and make products using it under the terms of
// the CERN-OHL-P v2 (https:/cern.ch/cern-ohl). This documentation is distributed WITHOUT ANY
// EXPRESS OR IMPLIED WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A
// PARTICULAR PURPOSE. Please see the CERN-OHL-P v2 for applicable conditions.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
module lfsr # (
		parameter CLOCK_POL = 1,
		parameter RESET_POL = 0,
		parameter ARESET = 1,
		parameter LFSR_WIDTH = 8,
		parameter [LFSR_WIDTH-1:0] POLYNOMIAL = 8'b10111000,	// x^8 + x^6 + x^5 + x^4 + 1
		parameter [LFSR_WIDTH-1:0] INITIAL_VAL = {LFSR_WIDTH{1'b1}}
	) (
		input CLOCK,
		input RESET,
		input IN_CLR,
		input IN_ENA,
		output [LFSR_WIDTH-1:0] OUT_DATA,
		output OUT_ENA,
		output OUT_CYCLE
	);

	reg [LFSR_WIDTH-1:0] data_reg;
	reg ena_reg;

	wire clk, rst, a_rst, s_rst;

	wire [LFSR_WIDTH-1:0] next_data;

	assign clk = (CLOCK_POL == 0)? ~CLOCK : CLOCK;
	assign rst = (RESET_POL == 0)? ~RESET : RESET;
	assign a_rst = (ARESET != 0)? rst : 1'b0;
	assign s_rst = (ARESET == 0)? rst : 1'b0;
	
	assign next_data = {data_reg[LFSR_WIDTH-2:0], ^(data_reg & POLYNOMIAL)};

	assign OUT_DATA = data_reg;
	assign OUT_ENA = ena_reg;
	assign OUT_CYCLE = (next_data == INITIAL_VAL);

	always @(posedge clk or posedge a_rst) begin
		if (a_rst)
			data_reg <= {LFSR_WIDTH{1'b0}};
		else if (s_rst)
			data_reg <= {LFSR_WIDTH{1'b0}};
		else begin
			if (IN_CLR)
				data_reg <= {LFSR_WIDTH{1'b0}};
			else if (IN_ENA)
				if (data_reg == {LFSR_WIDTH{1'b0}})
					data_reg <= INITIAL_VAL;
				else if (next_data == INITIAL_VAL)
					data_reg <= {LFSR_WIDTH{1'b0}};
				else
					data_reg <= next_data;
		end
	end

	always @(posedge clk or posedge a_rst) begin
		if (a_rst)
			ena_reg <= 1'b0;
		else if (s_rst)
			ena_reg <= 1'b0;
		else begin
			if ((!IN_CLR) && (IN_ENA))
				ena_reg <= 1'b1;
			else
				ena_reg <= 1'b0;
		end
	end

endmodule
