// The CERN-OHL-P v2 License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// This source describes Open Hardware and is licensed under the CERN-OHL-P v2.
// You may redistribute and modify this documentation and make products using it under the terms of
// the CERN-OHL-P v2 (https:/cern.ch/cern-ohl). This documentation is distributed WITHOUT ANY
// EXPRESS OR IMPLIED WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A
// PARTICULAR PURPOSE. Please see the CERN-OHL-P v2 for applicable conditions.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
module lvds_loopback # (
		parameter DATA_WIDTH = 8,
		parameter LFSR_WIDTH = 8,
		parameter [LFSR_WIDTH-1:0] POLYNOMIAL = 8'b10111000,	// x^8 + x^6 + x^5 + x^4 + 1
		parameter [LFSR_WIDTH-1:0] INITIAL_VAL = {LFSR_WIDTH{1'b1}},
		parameter RX_DATA_INV = 0
	) (
		input TX_CLK,
		input TX_PLL_LOCKED,
		input RX_CLK,
		input RX_PLL_LOCKED,
		input IN_RUN,
		input [DATA_WIDTH-1:0] LVDS_RX_DATA,
		input [DATA_WIDTH-1:0] LVDS_RX_INV_DATA,
		output [DATA_WIDTH-1:0] LVDS_TX_DATA,
		output OUT_RX_EXISTS,
		output OUT_ERROR
	);

	reg rx_exists_reg;
	reg error_reg;

	wire tx_run;
	wire rx_run;
	wire [LFSR_WIDTH-1:0] lfsr_data;
	wire [LFSR_WIDTH-1:0] comp_data;
	wire comp_data_ena;
	wire [DATA_WIDTH-1:0] rx_data;
	wire [DATA_WIDTH-1:0] rx_inv_data;

	assign LVDS_TX_DATA = lfsr_data[DATA_WIDTH-1:0];

	assign OUT_RX_EXISTS = rx_exists_reg;
	assign OUT_ERROR = error_reg;

	s_ff # (
		.CLOCK_POL(1),
		.RESET_POL(0),
		.ARESET(1),
		.DATA_WIDTH(1),
		.SHIFT_NUMS(2)
	) s_ff_u1 (
		.CLOCK(TX_CLK), 
		.RESET(TX_PLL_LOCKED), 
		.IN_DATA(IN_RUN),
		.OUT_DATA(tx_run)
	);

	lfsr # (
		.CLOCK_POL(1),
		.RESET_POL(0),
		.ARESET(1),
		.LFSR_WIDTH(LFSR_WIDTH),
		.POLYNOMIAL(POLYNOMIAL),
		.INITIAL_VAL(INITIAL_VAL)
	) lfsr_u1 (
		.CLOCK(TX_CLK),
		.RESET(TX_PLL_LOCKED),
		.IN_CLR(~tx_run),
		.IN_ENA(tx_run),
		.OUT_DATA(lfsr_data),
		.OUT_ENA(),
		.OUT_CYCLE()
	);

	s_ff # (
		.CLOCK_POL(1),
		.RESET_POL(0),
		.ARESET(1),
		.DATA_WIDTH(1),
		.SHIFT_NUMS(2)
	) s_ff_u2 (
		.CLOCK(RX_CLK), 
		.RESET(RX_PLL_LOCKED), 
		.IN_DATA(IN_RUN),
		.OUT_DATA(rx_run)
	);

	always @(posedge RX_CLK or negedge RX_PLL_LOCKED) begin
		if (!RX_PLL_LOCKED)
			rx_exists_reg <= 1'b0;
		else begin
			if (rx_run == 1'b0)
				rx_exists_reg <= 1'b0;
			if (LVDS_RX_DATA != {DATA_WIDTH{1'b0}})
				rx_exists_reg <= 1'b1;
		end
	end

	lfsr # (
		.CLOCK_POL(1),
		.RESET_POL(0),
		.ARESET(1),
		.LFSR_WIDTH(LFSR_WIDTH),
		.POLYNOMIAL(POLYNOMIAL),
		.INITIAL_VAL(INITIAL_VAL)
	) lfsr_u2 (
		.CLOCK(RX_CLK),
		.RESET(RX_PLL_LOCKED),
		.IN_CLR(~rx_run),
		.IN_ENA(rx_exists_reg),
		.OUT_DATA(comp_data),
		.OUT_ENA(comp_data_ena),
		.OUT_CYCLE()
	);

	s_ff #(
		.CLOCK_POL(1),
		.RESET_POL(0),
		.ARESET(1),
		.DATA_WIDTH(DATA_WIDTH),
		.SHIFT_NUMS(2)
	) s_ff_u3 (
		.CLOCK(RX_CLK), 
		.RESET(RX_PLL_LOCKED), 
		.IN_DATA(LVDS_RX_DATA),
		.OUT_DATA(rx_data)
	);

	generate
		if (RX_DATA_INV) begin
			s_ff #(
				.CLOCK_POL(1),
				.RESET_POL(0),
				.ARESET(1),
				.DATA_WIDTH(DATA_WIDTH),
				.SHIFT_NUMS(2)
			) s_ff_u4 (
				.CLOCK(RX_CLK), 
				.RESET(RX_PLL_LOCKED), 
				.IN_DATA(LVDS_RX_INV_DATA),
				.OUT_DATA(rx_inv_data)
			);
		end
	endgenerate

	always @(posedge RX_CLK or negedge RX_PLL_LOCKED) begin
		if (!RX_PLL_LOCKED)
			error_reg <= 1'b0;
		else begin
			if (comp_data_ena) begin
				if ((comp_data[DATA_WIDTH-1:0] != rx_data) || ((RX_DATA_INV) && (~comp_data[DATA_WIDTH-1:0] != rx_inv_data)))
					error_reg = 1'b1;
			end
			else
				error_reg = 1'b0;
		end
	end

endmodule
