// The CERN-OHL-P v2 License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// This source describes Open Hardware and is licensed under the CERN-OHL-P v2.
// You may redistribute and modify this documentation and make products using it under the terms of
// the CERN-OHL-P v2 (https:/cern.ch/cern-ohl). This documentation is distributed WITHOUT ANY
// EXPRESS OR IMPLIED WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A
// PARTICULAR PURPOSE. Please see the CERN-OHL-P v2 for applicable conditions.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
`include "../../../Efinity/2023.2/pt/sim_models/verilog/EFX_PLL_V2.v"
`include "../../../Efinity/2023.2/pt/sim_models/verilog/EFX_LVDS_TX_V1.v"
`include "../../../Efinity/2023.2/pt/sim_models/verilog/EFX_LVDS_RX_V1.v"

`timescale 1ps / 1ps

module sim;

	// Parameters
	parameter CLK_CYCLE = 40000;							// 25MHz
	parameter LFSR_WIDTH = 8;
	parameter [LFSR_WIDTH-1:0] POLYNOMIAL = 8'b10111000;	// x^8 + x^6 + x^5 + x^4 + 1
	parameter [LFSR_WIDTH-1:0] INITIAL_VAL = {LFSR_WIDTH{1'b1}};
//	parameter RX_DATA_INV = 0;
	parameter RX_DATA_INV = 1;

	parameter SERIALIZATION = 8;		// 1 to 8

	parameter ENABLE_SERIALIZATION = (SERIALIZATION >= 2);
	parameter DATA_WIDTH = SERIALIZATION;

	parameter TX_PLL_M = SERIALIZATION * 8;
	parameter TX_RX_PLL_D = SERIALIZATION;

	reg CLOCK = 1'b0;

	// Inputs
	reg TXPLL_RST_N = 1'b0;
	reg RXPLL_RST_N = 1'b0;

	reg IN_RUN = 1'b0;

	reg [DATA_WIDTH-1:0] LVDS_RX_DATA_JAM = {DATA_WIDTH{1'b0}};
	reg [DATA_WIDTH-1:0] LVDS_RX_INV_DATA_JAM = {DATA_WIDTH{1'b0}};

	// Outputs
	wire OUT_RX_EXISTS;
	wire OUT_ERROR;

	// External loopback
	wire LVDS_CLK_P, LVDS_CLK_N;
	wire LVDS_DATA_P, LVDS_DATA_N;

	// inter-module connection
	wire TXPLL_SCLK;
	wire TXPLL_FCLK;
	wire TXPLL_LOCKED;
	wire RXPLL_SCLK;
	wire RXPLL_FCLK;
	wire RXPLL_LOCKED;
	wire [7:0] LVDS_RX_CLK;
	wire [7:0] LVDS_RX_DATA;
	wire [7:0] LVDS_RX_INV_DATA;
	wire [7:0] LVDS_TX_DATA;

	if (DATA_WIDTH < 8)
		assign LVDS_TX_DATA[7:DATA_WIDTH] = {(8-DATA_WIDTH){1'b0}};

	EFX_PLL_V2 # (
		.M(TX_PLL_M),
		.N(1),
		.O(2),
		.CLKOUT0_DIV(2),
		.CLKOUT0_PHASE(90),
		.CLKOUT1_DIV(TX_RX_PLL_D),
		.CLKOUT1_PHASE(0),
		.CLKOUT2_DIV(1),
		.CLKOUT2_PHASE(0),
   		.FEEDBACK_CLK("INTERNAL")
	) eut_efx_pll_v2_tx (
		.CLKOUT0(TXPLL_FCLK),
		.CLKOUT1(TXPLL_SCLK),
		.CLKOUT2(),
		.LOCKED(TXPLL_LOCKED),
		.RSTN(TXPLL_RST_N),
		.CLKIN({3'b000, CLOCK}),
		.CLKSEL(2'b00),
		.FBK(1'b0)
	);

	EFX_LVDS_TX_V1 # (
		.ENABLE_SERIALIZATION(1),
		.SERIALIZATION((ENABLE_SERIALIZATION)? SERIALIZATION : 8),
		.MODE("CLKOUT"),
		.DIV2_CLKOUT(0)
	) eut_efx_lvds_tx_v1_clk (
		.RST(1'b0),
		.FASTCLK(TXPLL_FCLK),
		.SLOWCLK(TXPLL_SCLK),
		.OE(1'b1),
		.O(8'h00),
		.P(LVDS_CLK_P),
		.N(LVDS_CLK_N)
	);

	EFX_LVDS_TX_V1 # (
		.ENABLE_SERIALIZATION(ENABLE_SERIALIZATION),
		.SERIALIZATION((ENABLE_SERIALIZATION)? SERIALIZATION : 8),
		.MODE("DATA"),
		.DIV2_CLKOUT(0)
	) eut_efx_lvds_tx_v1_data (
		.RST(1'b0),
		.FASTCLK(TXPLL_FCLK),
		.SLOWCLK(TXPLL_SCLK),
		.OE(1'b1),
		.O(LVDS_TX_DATA),
		.P(LVDS_DATA_P),
		.N(LVDS_DATA_N)
	);

	EFX_PLL_V2 # (
		.M(2),
		.N(2),
		.O(4),
		.CLKOUT0_DIV(TX_RX_PLL_D),
		.CLKOUT0_PHASE(0),
		.CLKOUT1_DIV(1),
		.CLKOUT1_PHASE(0),
		.CLKOUT2_DIV(2),
		.CLKOUT2_PHASE(90),
   		.FEEDBACK_CLK("CLK0")
	) eut_efx_pll_v2_rx (
		.CLKOUT0(RXPLL_SCLK),
		.CLKOUT1(),
		.CLKOUT2(RXPLL_FCLK),
		.LOCKED(RXPLL_LOCKED),
		.RSTN(RXPLL_RST_N),
		.CLKIN({3'b000, LVDS_RX_CLK[0]}),
		.CLKSEL(2'b00),
		.FBK(RXPLL_SCLK)
	);

	EFX_LVDS_RX_V1 # (
		.ENABLE_DESERIALIZATION(0),
		.DESERIALIZATION(8)
	) eut_efx_lvds_rx_v1_clk (
		.RST(1'b0),
		.FASTCLK(1'b0),
		.SLOWCLK(1'b0),
		.P(LVDS_CLK_P),
		.N(LVDS_CLK_N),
		.I(LVDS_RX_CLK)
	);

	EFX_LVDS_RX_V1 # (
		.ENABLE_DESERIALIZATION(ENABLE_SERIALIZATION),
		.DESERIALIZATION(SERIALIZATION)
	) eut_efx_lvds_rx_v1_data (
		.RST(1'b0),
		.FASTCLK(RXPLL_FCLK),
		.SLOWCLK(RXPLL_SCLK),
		.P(LVDS_DATA_P),
		.N(LVDS_DATA_N),
		.I(LVDS_RX_DATA)
	);

	EFX_LVDS_RX_V1 # (
		.ENABLE_DESERIALIZATION(ENABLE_SERIALIZATION),
		.DESERIALIZATION(SERIALIZATION)
	) eut_efx_lvds_rx_v1_inv_data (
		.RST(1'b0),
		.FASTCLK(RXPLL_FCLK),
		.SLOWCLK(RXPLL_SCLK),
		.P(LVDS_DATA_N),
		.N(LVDS_DATA_P),
		.I(LVDS_RX_INV_DATA)
	);

	lvds_loopback # (
		.DATA_WIDTH(DATA_WIDTH),
		.LFSR_WIDTH(LFSR_WIDTH),
		.POLYNOMIAL(POLYNOMIAL),
		.INITIAL_VAL(INITIAL_VAL),
		.RX_DATA_INV(RX_DATA_INV)
	) dut1 (
		.TX_CLK(TXPLL_SCLK),
		.TX_PLL_LOCKED(TXPLL_LOCKED),
		.RX_CLK((ENABLE_SERIALIZATION)? RXPLL_SCLK : ~RXPLL_SCLK),
		.RX_PLL_LOCKED(RXPLL_LOCKED),
		.IN_RUN(IN_RUN),
		.LVDS_RX_DATA(LVDS_RX_DATA[DATA_WIDTH-1:0] + LVDS_RX_DATA_JAM[DATA_WIDTH-1:0]),
		.LVDS_RX_INV_DATA(LVDS_RX_INV_DATA[DATA_WIDTH-1:0] + LVDS_RX_INV_DATA_JAM[DATA_WIDTH-1:0]),
		.LVDS_TX_DATA(LVDS_TX_DATA[DATA_WIDTH-1:0]),
		.OUT_RX_EXISTS(OUT_RX_EXISTS),
		.OUT_ERROR(OUT_ERROR)
	);

	initial begin
		$dumpfile("outflow/lvds_loopback.vcd");
		$dumpvars(0, sim);
	end

	always #(CLK_CYCLE / 2) CLOCK = ~CLOCK;

	task task_delay (
			input integer value
		);
		begin
			repeat (value)
				@(posedge CLOCK);
			#1;
		end
	endtask

	initial begin
		task_delay(10);

		TXPLL_RST_N = 1'b1;							task_delay(10);
		RXPLL_RST_N = 1'b1;							task_delay(100);

		IN_RUN = 1'b1;								task_delay(1000);
		LVDS_RX_DATA_JAM[0] = 1'b1;					task_delay(1);
		LVDS_RX_DATA_JAM[0] = 1'b0;					task_delay(10);
		IN_RUN = 1'b0;								task_delay(100);
		IN_RUN = 1'b1;								task_delay(1000);
		LVDS_RX_INV_DATA_JAM[DATA_WIDTH-1] = 1'b1;	task_delay(1);
		LVDS_RX_INV_DATA_JAM[DATA_WIDTH-1] = 1'b0;	task_delay(10);
		IN_RUN = 1'b0;								task_delay(100);

		$finish;
	end

endmodule
