// The CERN-OHL-P v2 License
//
// Copyright (c) Since 2024 RIGHTxLIGHT Co., Ltd.
//
// This source describes Open Hardware and is licensed under the CERN-OHL-P v2.
// You may redistribute and modify this documentation and make products using it under the terms of
// the CERN-OHL-P v2 (https:/cern.ch/cern-ohl). This documentation is distributed WITHOUT ANY
// EXPRESS OR IMPLIED WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A
// PARTICULAR PURPOSE. Please see the CERN-OHL-P v2 for applicable conditions.
//
// Revisions:
// Rev.1.0		2024/04/16		Initial release
//
// -------------------------------------------------------------------------------------------------
module sled_drv # (
		parameter CLOCK_POL = 1,
		parameter RESET_POL = 0,
		parameter ARESET = 1,
																// WS2812S parameters
		parameter T_WIDTH = 125,								// 1.25us * 100MHz = 125
		parameter T0H_WIDTH = 40,								// 0.4us * 100MHz = 40
		parameter T1H_WIDTH = 85,								// 0.85us * 100MHz = 85
		parameter T_RESET_WIDTH = 51000,						// 51us * 100MHz / T_WIDTH / DATA_WIDTH = 51000
		parameter LED_NUMS = 4,
		parameter ADDR_WIDTH = 2,								// Max 7bit	128LEDs
		parameter DATA_WIDTH = 24								// Fixed 24bit
	) (
		input CLOCK,
		input RESET,
		input [DATA_WIDTH-1:0] IN_DATA,
		input [ADDR_WIDTH-1:0] IN_ADDR,
		input IN_WRITE,
		input IN_LED_SET,
		output OUT_BUSY,
		output OUT_LED_DATA
	);

	localparam TX_CT_WIDTH = $clog2(T_WIDTH+1);
	localparam DATA_CT_WIDTH = $clog2(DATA_WIDTH+1);
	localparam LED_CT_WIDTH = $clog2(LED_NUMS+1);
	localparam RESET_CT_WIDTH = $clog2(T_RESET_WIDTH+1);
	localparam [DATA_CT_WIDTH-1:0] DATA_CT_END = DATA_WIDTH - 1;
	localparam [LED_CT_WIDTH-1:0] LED_CT_END = LED_NUMS - 1;
	localparam [RESET_CT_WIDTH-1:0] RST_CT_END = T_RESET_WIDTH - 1;

	localparam [1:0] READY_STATE = 2'd0;
	localparam [1:0] TRANSFER_STATE = 2'd1;
	localparam [1:0] RESET_STATE = 2'd2;

	reg [1:0] sled_state_reg;

	reg [DATA_WIDTH-1:0] led_data_arr [0:LED_NUMS-1];
	reg [DATA_WIDTH-1:0] led_data_reg;
	reg led_set_out_reg;

	wire clk, rst, a_rst, s_rst;
	wire [DATA_WIDTH-1:0] wr_data;
	wire led_set_fall, led_set_end, led_rst_end;
	wire tx_carry, data_carry;
	wire [TX_CT_WIDTH-1:0] tx_ct;
	wire [DATA_CT_WIDTH-1:0] data_count;
	wire [LED_CT_WIDTH-1:0] led_count;
	wire data_end, led_end;

	assign clk = (CLOCK_POL == 0)? ~CLOCK : CLOCK;
	assign rst = (RESET_POL == 0)? ~RESET : RESET;
	assign a_rst = (ARESET != 0)? rst : 1'b0;
	assign s_rst = (ARESET == 0)? rst : 1'b0;
	assign led_set_end = (tx_carry) && (data_end) && (led_end);

	assign OUT_BUSY = sled_state_reg != READY_STATE;

	generate
		for (genvar i = 0; i < DATA_WIDTH; i = i + 1) begin
			assign wr_data[DATA_WIDTH-1-i] = IN_DATA[i];			
		end
	endgenerate

	integer a;
	initial begin
		for (a = 0; a < LED_NUMS; a = a + 1)
			led_data_arr[a] = {DATA_WIDTH{1'b0}};
	end

	always @(posedge clk) begin
		if (IN_WRITE)
			led_data_arr[IN_ADDR] = wr_data;
		led_data_reg = led_data_arr[led_count];
	end

	edge_det # (
		.CLOCK_POL(CLOCK_POL),
		.RESET_POL(RESET_POL),
		.ARESET(ARESET)
	) edge_det_u1 (
		.CLOCK(CLOCK), 
		.RESET(RESET), 
		.IN_DATA(IN_LED_SET),
		.OUT_RISE(),
		.OUT_FALL(led_set_fall)
	);

	always @(posedge clk or posedge a_rst) begin
		if (a_rst)
			sled_state_reg <= READY_STATE;
		else if (s_rst)
			sled_state_reg <= READY_STATE;
	else begin
			if (led_set_fall)
				sled_state_reg <= TRANSFER_STATE;
			else if (led_set_end)
				sled_state_reg <= RESET_STATE;
			else if (led_rst_end)
				sled_state_reg <= READY_STATE;
		end
	end

	counter # (
		.CLOCK_POL(CLOCK_POL),
		.RESET_POL(RESET_POL),
		.ARESET(ARESET),
		.CT_WIDTH(TX_CT_WIDTH),
		.CONTINUE(1),
		.CARRY_START(0)
	) counter_u1 (
		.CLOCK(CLOCK), 
		.RESET(RESET), 
		.IN_COUNT(T_WIDTH[TX_CT_WIDTH-1:0]), 
		.IN_COUNT_CLR(led_set_end), 
		.IN_COUNT_RUN(sled_state_reg == TRANSFER_STATE), 
		.IN_COUNT_STB(1'b1), 
		.IN_COMP_LOW({{TX_CT_WIDTH-1{1'b0}}, 1'b1}), 
		.IN_COMP_HIGH((led_data_reg[data_count])? T1H_WIDTH[TX_CT_WIDTH-1:0] : T0H_WIDTH[TX_CT_WIDTH-1:0]), 
		.OUT_COUNT(tx_ct), 
		.OUT_CARRY(tx_carry),
		.OUT_COMP(OUT_LED_DATA),
		.OUT_COUNT_STB()
	);

	counter # (
		.CLOCK_POL(CLOCK_POL),
		.RESET_POL(RESET_POL),
		.ARESET(ARESET),
		.CT_WIDTH(DATA_CT_WIDTH),
		.CONTINUE(1),
		.CARRY_START(0)
	) counter_u2 (
		.CLOCK(CLOCK), 
		.RESET(RESET), 
		.IN_COUNT(DATA_WIDTH[DATA_CT_WIDTH-1:0]), 
		.IN_COUNT_CLR(led_set_end), 
		.IN_COUNT_RUN(sled_state_reg == TRANSFER_STATE), 
		.IN_COUNT_STB(tx_carry), 
		.IN_COMP_LOW(DATA_CT_END), 
		.IN_COMP_HIGH(DATA_CT_END), 
		.OUT_COUNT(data_count), 
		.OUT_CARRY(data_carry),
		.OUT_COMP(data_end),
		.OUT_COUNT_STB()
	);

	counter # (
		.CLOCK_POL(CLOCK_POL),
		.RESET_POL(RESET_POL),
		.ARESET(ARESET),
		.CT_WIDTH(LED_CT_WIDTH),
		.CONTINUE(0),
		.CARRY_START(0)
	) counter_u3 (
		.CLOCK(CLOCK), 
		.RESET(RESET), 
		.IN_COUNT(LED_NUMS[LED_CT_WIDTH-1:0]), 
		.IN_COUNT_CLR(led_set_end), 
		.IN_COUNT_RUN(sled_state_reg == TRANSFER_STATE), 
		.IN_COUNT_STB(data_carry), 
		.IN_COMP_LOW(LED_CT_END), 
		.IN_COMP_HIGH(LED_CT_END), 
		.OUT_COUNT(led_count), 
		.OUT_CARRY(),
		.OUT_COMP(led_end),
		.OUT_COUNT_STB()
	);

	counter # (
		.CLOCK_POL(CLOCK_POL),
		.RESET_POL(RESET_POL),
		.ARESET(ARESET),
		.CT_WIDTH(RESET_CT_WIDTH),
		.CONTINUE(0),
		.CARRY_START(0)
	) counter_u4 (
		.CLOCK(CLOCK), 
		.RESET(RESET), 
		.IN_COUNT(T_RESET_WIDTH[RESET_CT_WIDTH-1:0]), 
		.IN_COUNT_CLR(led_rst_end), 
		.IN_COUNT_RUN(sled_state_reg == RESET_STATE), 
		.IN_COUNT_STB(1'b1), 
		.IN_COMP_LOW(RST_CT_END), 
		.IN_COMP_HIGH(RST_CT_END), 
		.OUT_COUNT(), 
		.OUT_CARRY(),
		.OUT_COMP(led_rst_end),
		.OUT_COUNT_STB()
	);

endmodule
